# Copyright 2015 - StackStorm, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.db.v2 import api as db_api
from mistral.services import workflows as wf_service
from mistral.tests.unit.engine import base
from mistral.workflow import states


class WorkflowStopTest(base.EngineTestCase):
    def setUp(self):
        super(WorkflowStopTest, self).setUp()

        WORKFLOW = """
        version: '2.0'

        wf:
          type: direct
          tasks:
            task1:
              action: std.echo output="Echo"
              on-complete:
                - task2

            task2:
              action: std.echo output="foo"
              wait-before: 3
        """
        wf_service.create_workflows(WORKFLOW)

        self.exec_id = self.engine.start_workflow('wf').id

    def test_stop_failed(self):
        self.engine.stop_workflow(self.exec_id, states.SUCCESS, "Force stop")

        self.await_workflow_success(self.exec_id)

        wf_ex = db_api.get_workflow_execution(self.exec_id)

        self.assertEqual(states.SUCCESS, wf_ex.state)
        self.assertEqual("Force stop", wf_ex.state_info)

    def test_stop_succeeded(self):
        self.engine.stop_workflow(self.exec_id, states.ERROR, "Failure")

        self.await_workflow_error(self.exec_id)

        wf_ex = db_api.get_workflow_execution(self.exec_id)

        self.assertEqual(states.ERROR, wf_ex.state)
        self.assertEqual("Failure", wf_ex.state_info)



# Copyright 2017 - Brocade Communications Systems, Inc.
# Copyright 2018 - Extreme Networks, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging
from stevedore import exception as sd_exc

from mistral.executors import base as exe
from mistral.executors import default_executor as d
from mistral.executors import remote_executor as r
from mistral.tests.unit.executors import base


LOG = logging.getLogger(__name__)


class PluginTest(base.ExecutorTestCase):

    def tearDown(self):
        exe.cleanup()
        super(PluginTest, self).tearDown()

    def test_get_local_executor(self):
        executor = exe.get_executor('local')

        self.assertIsInstance(executor, d.DefaultExecutor)

    def test_get_remote_executor(self):
        executor = exe.get_executor('remote')

        self.assertIsInstance(executor, r.RemoteExecutor)

    def test_get_bad_executor(self):
        self.assertRaises(sd_exc.NoMatches, exe.get_executor, 'foobar')



# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
#   This module implements SQLAlchemy-based types for dict and list
#   expressed by json-strings
#

import sqlalchemy as sa
from sqlalchemy.dialects import mysql
from sqlalchemy.ext import mutable

from mistral import utils


class JsonEncoded(sa.TypeDecorator):
    """Represents an immutable structure as a json-encoded string."""

    impl = sa.Text

    def process_bind_param(self, value, dialect):
        return utils.to_json_str(value)

    def process_result_value(self, value, dialect):
        return utils.from_json_str(value)


class MutableList(mutable.Mutable, list):
    @classmethod
    def coerce(cls, key, value):
        """Convert plain lists to MutableList."""
        if not isinstance(value, MutableList):
            if isinstance(value, list):
                return MutableList(value)

            # this call will raise ValueError
            return mutable.Mutable.coerce(key, value)
        return value

    def __add__(self, value):
        """Detect list add events and emit change events."""
        list.__add__(self, value)
        self.changed()

    def append(self, value):
        """Detect list add events and emit change events."""
        list.append(self, value)
        self.changed()

    def __setitem__(self, key, value):
        """Detect list set events and emit change events."""
        list.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, i):
        """Detect list del events and emit change events."""
        list.__delitem__(self, i)
        self.changed()


def JsonDictType():
    """Returns an SQLAlchemy Column Type suitable to store a Json dict."""
    return mutable.MutableDict.as_mutable(JsonEncoded)


def JsonListType():
    """Returns an SQLAlchemy Column Type suitable to store a Json array."""
    return MutableList.as_mutable(JsonEncoded)


def MediumText():
    # TODO(rakhmerov): Need to do for postgres.
    return sa.Text().with_variant(mysql.MEDIUMTEXT(), 'mysql')


class JsonEncodedMediumText(JsonEncoded):
    impl = MediumText()


def JsonMediumDictType():
    return mutable.MutableDict.as_mutable(JsonEncodedMediumText)


def LongText():
    # TODO(rakhmerov): Need to do for postgres.
    return sa.Text().with_variant(mysql.LONGTEXT(), 'mysql')


class JsonEncodedLongText(JsonEncoded):
    impl = LongText()


def JsonLongDictType():
    return mutable.MutableDict.as_mutable(JsonEncodedLongText)



# Copyright 2019 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Create scheduled jobs table.

Revision ID: 034
Revises: 033
Create Date: 2019-07-01 17:38:41.153354

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.engine import reflection

from mistral.db.sqlalchemy import types as st

# revision identifiers, used by Alembic.
revision = '034'
down_revision = '033'


def upgrade():
    # NOTE(rakhmerov): We have to check if the table already
    # exists and drop it, if needed. This is because the DB
    # model for scheduled jobs was released w/o a migration
    # in the first place, so for some users the table was
    # created automatically at Mistral run based on the model.
    # But the structure of the table is old so we need to
    # recreate it anyway in this migration. It's safe to drop
    # this table because it contains temporary data.
    inspect = reflection.Inspector.from_engine(op.get_bind())

    if 'scheduled_jobs_v2' in inspect.get_table_names():
        op.drop_table('scheduled_jobs_v2')

    op.create_table(
        'scheduled_jobs_v2',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('run_after', sa.Integer(), nullable=True),
        sa.Column(
            'target_factory_func_name',
            sa.String(length=200),
            nullable=True
        ),
        sa.Column('func_name', sa.String(length=80), nullable=True),
        sa.Column('func_args', st.JsonEncoded(), nullable=True),
        sa.Column('func_arg_serializers', st.JsonEncoded(), nullable=True),
        sa.Column('auth_ctx', st.JsonEncoded(), nullable=True),
        sa.Column('execute_at', sa.DateTime(), nullable=False),
        sa.Column('captured_at', sa.DateTime(), nullable=True),
        sa.Column('key', sa.String(length=250), nullable=True),

        sa.PrimaryKeyConstraint('id'),
    )



# Copyright 2018 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""workbook_empty_namespace

Revision ID: 029
Revises: 028
Create Date: 2018-08-08 10:30:00.727769

"""

# revision identifiers, used by Alembic.
revision = '029'
down_revision = '028'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import table, column


def upgrade():
    wb_def = table('workbooks_v2', column('namespace'))
    session = sa.orm.Session(bind=op.get_bind())
    with session.begin(subtransactions=True):
        session.execute(
            wb_def.update().values(namespace='').where(
                wb_def.c.namespace == None))  # noqa

    session.commit()



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import eventlet
from oslo_log import log as logging

from mistral import context as auth_ctx
from mistral.db.v2 import api as db_api
from mistral.engine import post_tx_queue
from mistral.engine import workflow_handler
from mistral import exceptions
from mistral.scheduler import base as sched_base
from mistral.workflow import states

PAUSING = 'PAUSING'
PAUSED = 'PAUSED'
RUNNING = 'RUNNING'

_VALID_TRANSITIONS = {
    PAUSING: [PAUSED],
    PAUSED: [PAUSED, RUNNING],
    RUNNING: [PAUSED, RUNNING]
}

_ALL_STATES = [
    PAUSING,
    PAUSED,
    RUNNING
]

LOG = logging.getLogger(__name__)

_PAUSE_EXECUTIONS_PATH = 'mistral.services.maintenance._pause_executions'
_PAUSE_EXECUTION_PATH = 'mistral.services.maintenance._pause_execution'
_RESUME_EXECUTIONS_PATH = 'mistral.services.maintenance._resume_executions'
_RESUME_EXECUTION_PATH = 'mistral.services.maintenance._resume_execution'
_AWAIT_PAUSE_EXECUTION_PATH = \
    'mistral.services.maintenance.await_pause_executions'


def is_valid_transition(old_state, new_state):
    return new_state in _VALID_TRANSITIONS.get(old_state, [])


def pause_running_executions(skip_tx=False):
    execution_ids = [(ex.id, ex.project_id) for ex in
                     db_api.get_workflow_executions(state=states.RUNNING,
                     insecure=True)]

    LOG.info("Number of find workflow executions is {}",
             len(execution_ids))

    if skip_tx:
        sched = sched_base.get_system_scheduler()
        for wf_ex_id, project_id in execution_ids:
            job = sched_base.SchedulerJob(
                func_name=_PAUSE_EXECUTION_PATH,
                func_args={
                    'wf_ex_id': wf_ex_id,
                    'project_id': project_id
                }
            )
            sched.schedule(job)
        return

    for wf_ex_id, project_id in execution_ids:
        try:
            with db_api.transaction(skip=skip_tx):
                _pause_execution(wf_ex_id, project_id)
        except BaseException as e:
            LOG.error(str(e))

    return True


def _pause_execution(wf_ex_id, project_id, skip_tx=False):
    auth_ctx.set_ctx(
        auth_ctx.MistralContext(
            user=None,
            auth_token=None,
            is_admin=True
        )
    )

    current_state = db_api.get_maintenance_status()

    if current_state != PAUSING:
        return False

    wf_ex = db_api.get_workflow_execution(wf_ex_id)

    if wf_ex.root_execution_id:
        trace_uuid = wf_ex.root_execution_id
    else:
        trace_uuid = wf_ex.id

    auth_ctx.set_ctx(
        auth_ctx.MistralContext(
            tenant=project_id,
            trace_uuid=trace_uuid
        )
    )

    if states.is_running(wf_ex.state):
        workflow_handler.pause_workflow(wf_ex)
        LOG.info('Execution {} was paused', wf_ex_id)


def await_pause_executions(skip_tx=False):
    auth_ctx.set_ctx(
        auth_ctx.MistralContext(
            user=None,
            auth_token=None,
            is_admin=True
        )
    )

    if skip_tx:
        current_state = db_api.get_maintenance_status()

        if current_state != PAUSING:
            return False

        tasks = db_api.get_task_executions(
            state=states.RUNNING, insecure=True
        )

        if not tasks:
            if db_api.get_maintenance_status() == PAUSING:
                db_api.update_maintenance_status(PAUSED)
            return

        LOG.info('The following tasks have RUNNING state: {}', [
            task.id for task in tasks
        ])

        sched = sched_base.get_system_scheduler()
        job = sched_base.SchedulerJob(
            run_after=1,
            func_name=_AWAIT_PAUSE_EXECUTION_PATH,
            func_args={'skip_tx': True}
        )
        sched.schedule(job)
        return

    while True:
        with db_api.transaction(skip=skip_tx):
            current_state = db_api.get_maintenance_status()

            if current_state != PAUSING:
                return False

            tasks = db_api.get_task_executions(
                state=states.RUNNING, insecure=True
            )

            if not tasks:
                return True

            LOG.info('The following tasks have RUNNING state: {}', [
                task.id for task in tasks
            ])

            eventlet.sleep(1)


def change_maintenance_mode(new_state):
    if new_state not in _ALL_STATES:
        raise exceptions.MaintenanceException(
            'Not found {} maintenance state. List of states: {}'.format(
                new_state, _ALL_STATES
            )
        )

    if new_state == PAUSING:
        raise exceptions.MaintenanceException(
            'PAUSING is intermediate state. Consider PAUSED, RUNNING as new '
            'state')

    with db_api.transaction():
        current_state = db_api.get_maintenance_status()

        if current_state == new_state:
            LOG.info('State was already changed. Skip')
            return current_state

        sched = sched_base.get_system_scheduler()

        if new_state == PAUSED:
            job = sched_base.SchedulerJob(func_name=_PAUSE_EXECUTIONS_PATH)
            sched.schedule(job)
            db_api.update_maintenance_status(PAUSING)

            return PAUSING
        elif new_state == RUNNING:
            job = sched_base.SchedulerJob(func_name=_RESUME_EXECUTIONS_PATH)
            sched.schedule(job)
            db_api.update_maintenance_status(RUNNING)

            return RUNNING


@post_tx_queue.run
def _pause_executions(skip_tx=False):
    auth_ctx.set_ctx(
        auth_ctx.MistralContext(
            user=None,
            auth_token=None,
            is_admin=True
        )
    )

    if skip_tx:
        pause_running_executions(skip_tx)
        await_pause_executions(skip_tx)
        return

    if pause_running_executions() and await_pause_executions():
        with db_api.transaction():
            if db_api.get_maintenance_status() == PAUSING:
                db_api.update_maintenance_status(PAUSED)


@post_tx_queue.run
def _resume_executions(skip_tx=False):
    auth_ctx.set_ctx(
        auth_ctx.MistralContext(
            user=None,
            auth_token=None,
            is_admin=True
        )
    )
    sched = sched_base.get_system_scheduler()

    with db_api.transaction(skip=skip_tx):
        current_state = db_api.get_maintenance_status()

        if current_state != RUNNING:
            return

        paused_executions = db_api.get_workflow_executions(
            state=states.PAUSED, insecure=True
        )

        if not paused_executions:
            return

        for ex in paused_executions:
            if skip_tx:
                job = sched_base.SchedulerJob(
                    func_name=_RESUME_EXECUTION_PATH,
                    func_args={
                        'wf_ex_id': ex.id
                    }
                )
                sched.schedule(job)
            else:
                _resume_execution(wf_ex_id=ex.id)


def _resume_execution(wf_ex_id, skip_tx=False):
    wf_ex = db_api.get_workflow_execution(wf_ex_id)

    if wf_ex.root_execution_id:
        trace_uuid = wf_ex.root_execution_id
    else:
        trace_uuid = wf_ex.id

    auth_ctx.set_ctx(
        auth_ctx.MistralContext(
            tenant=wf_ex.project_id,
            trace_uuid=trace_uuid
        )
    )

    workflow_handler.resume_workflow(wf_ex)

    LOG.info('The following execution was resumed: {}', [wf_ex.id])



# Copyright 2018 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add started_at and finished_at to task execution

Revision ID: 031
Revises: 030
Create Date: 2018-10-03 20:09:45.582597

"""

# revision identifiers, used by Alembic.
revision = '031'
down_revision = '030'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'task_executions_v2',
        sa.Column('started_at', sa.DateTime(), nullable=True)
    )
    op.add_column(
        'task_executions_v2',
        sa.Column('finished_at', sa.DateTime(), nullable=True)
    )



# Copyright (c) 2016 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
from kombu.mixins import ConsumerMixin
import queue
import threading

from oslo_log import log as logging
from oslo_utils import eventletutils

from mistral.rpc.kombu import base as kombu_base

LOG = logging.getLogger(__name__)


class KombuRPCListener(ConsumerMixin):

    def __init__(self, connections, callback_queue):
        self._results = {}
        self._connections = itertools.cycle(connections)
        self._callback_queue = callback_queue
        self._thread = None
        self.connection = next(self._connections)

        self.ready = eventletutils.Event()

    def add_listener(self, correlation_id):
        self._results[correlation_id] = queue.Queue()

    def remove_listener(self, correlation_id):
        if correlation_id in self._results:
            del self._results[correlation_id]

    def get_consumers(self, Consumer, channel):
        consumers = [Consumer(
            self._callback_queue,
            callbacks=[self.on_message],
            accept=['pickle', 'json']
        )]
        self.ready.set()

        return consumers

    def start(self):
        if self._thread is None:
            self._thread = threading.Thread(target=self.run)
            self._thread.daemon = True
            self._thread.start()

    def on_message(self, response, message):
        """Callback on response.

        This method is automatically called when a response is incoming and
        decides if it is the message we are waiting for - the message with the
        result.

        :param response: the body of the amqp message already deserialized
            by kombu
        :param message: the plain amqp kombu.message with additional
            information
        """
        LOG.debug("Got response: {}", response)

        try:
            message.ack()
        except Exception as e:
            LOG.exception("Failed to acknowledge AMQP message: %s", e)
        else:
            LOG.debug("AMQP message acknowledged.")

            correlation_id = message.properties['correlation_id']

            queue = self._results.get(correlation_id)

            if queue:
                result = {
                    kombu_base.TYPE: 'error'
                    if message.properties.get('type') == 'error'
                    else None,
                    kombu_base.RESULT: response
                }

                queue.put(result)
            else:
                LOG.debug(
                    "Got a response, but seems like no process is waiting for "
                    "it [correlation_id={}]", correlation_id
                )

    def get_result(self, correlation_id, timeout):
        return self._results[correlation_id].get(block=True, timeout=timeout)

    def on_connection_error(self, exc, interval):
        self.ready.clear()

        self.connection = next(self._connections)

        LOG.debug("Broker connection failed: %s", exc)
        LOG.debug(
            "Sleeping for %s seconds, then retrying connection",
            interval
        )

    def wait_ready(self, timeout=10.0):
        """Waits for the listener to successfully declare the consumer

        :param timeout: timeout for waiting in seconds
        :return: same as :func:`~threading.Event.wait`
        :rtype: bool

        """
        if self.ready.wait(timeout=timeout):
            return self.connection
        else:
            return False



# Copyright 2019 - Nokia Networks, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.services import workflows as wf_service
from mistral.tests.unit.api import base
from mistral.tests.unit.engine import base as engine_base

WF_TEXT = """---
        version: "2.0"
        action_wf:
          tasks:
            action_task:
              action: std.noop
        fail_wf:
          tasks:
            fail_task:
              action: std.fail
        middle_wf:
          tasks:
            middle_task:
              workflow: action_wf
            fail_task:
              workflow: fail_wf
        main_wf:
          tasks:
            main_task:
              workflow: middle_wf
        """


class TestSubExecutionsController(base.APITest, engine_base.EngineTestCase):

    def setUp(self):
        super(TestSubExecutionsController, self).setUp()

        wf_service.create_workflows(WF_TEXT)

    def test_sub_executions_wf_ex_id(self):
        wf_ex = self.engine.start_workflow('main_wf')
        self.await_workflow_error(wf_ex.id)

        resp = self.app.get('/v2/executions/%s/executions' % wf_ex.id)

        self.assertEqual(200, resp.status_int)

        main_wf_ex_list = resp.json['executions']

        self.assertEqual(4, len(main_wf_ex_list))
        self._assert_single_item(main_wf_ex_list, workflow_name='main_wf')
        self._assert_single_item(main_wf_ex_list, workflow_name='action_wf')
        self._assert_single_item(main_wf_ex_list, workflow_name='fail_wf')

        middle_wf = self._assert_single_item(
            main_wf_ex_list,
            workflow_name='middle_wf'
        )

        # check the sub execs of a sub-ex
        resp = self.app.get('/v2/executions/%s/executions' % middle_wf['id'])

        self.assertEqual(200, resp.status_int)

        middle_wf_ex_list = resp.json['executions']

        self.assertEqual(3, len(middle_wf_ex_list))
        self._assert_single_item(middle_wf_ex_list, workflow_name='middle_wf')
        self._assert_single_item(middle_wf_ex_list, workflow_name='action_wf')
        self._assert_single_item(middle_wf_ex_list, workflow_name='fail_wf')

    def test_sub_executions_errors_only(self):
        wf_ex = self.engine.start_workflow('main_wf')
        self.await_workflow_error(wf_ex.id)

        resp = self.app.get(
            '/v2/executions/%s/executions?errors_only=True'
            % wf_ex.id
        )

        self.assertEqual(200, resp.status_int)

        main_wf_ex_list = resp.json['executions']

        self.assertEqual(3, len(main_wf_ex_list))
        self._assert_single_item(main_wf_ex_list, workflow_name='middle_wf')
        self._assert_single_item(main_wf_ex_list, workflow_name='fail_wf')
        self._assert_no_item(main_wf_ex_list, workflow_name='action_wf')

    def test_sub_executions_with_max_depth(self):
        wf_ex = self.engine.start_workflow('main_wf')
        self.await_workflow_error(wf_ex.id)

        resp = self.app.get(
            '/v2/executions/%s/executions?max_depth=1'
            % wf_ex.id
        )

        self.assertEqual(200, resp.status_int)

        main_wf_ex_list = resp.json['executions']

        self.assertEqual(2, len(main_wf_ex_list))
        self._assert_single_item(main_wf_ex_list, workflow_name='middle_wf')
        self._assert_single_item(main_wf_ex_list, workflow_name='main_wf')

    def test_sub_executions_task_id(self):
        wf_ex = self.engine.start_workflow('main_wf')
        self.await_workflow_error(wf_ex.id)

        resp = self.app.get('/v2/executions/%s/executions' % wf_ex.id)

        self.assertEqual(200, resp.status_int)

        main_wf_ex_list = resp.json['executions']

        self.assertEqual(4, len(main_wf_ex_list))
        middle_wf = self._assert_single_item(
            main_wf_ex_list,
            workflow_name='middle_wf'
        )

        resp = self.app.get(
            '/v2/tasks/%s/executions'
            % middle_wf['task_execution_id']
        )

        self.assertEqual(200, resp.status_int)

        main_task_ex_list = resp.json['executions']

        self.assertEqual(3, len(main_task_ex_list))
        self._assert_single_item(main_task_ex_list, workflow_name='fail_wf')
        self._assert_single_item(main_task_ex_list, workflow_name='middle_wf')
        self._assert_single_item(main_task_ex_list, workflow_name='action_wf')

    def test_sub_executions_with_include_output(self):
        wf_ex = self.engine.start_workflow('main_wf')
        self.await_workflow_error(wf_ex.id)

        resp = self.app.get(
            '/v2/executions/%s/executions?include_output=true'
            % wf_ex.id
        )

        self.assertEqual(200, resp.status_int)
        main_wf = self._assert_single_item(
            resp.json['executions'],
            workflow_name='main_wf'
        )

        self.assertIsNotNone(main_wf.get('output'))

        resp = self.app.get('/v2/executions/%s/executions' % wf_ex.id)

        self.assertEqual(200, resp.status_int)

        main_wf = self._assert_single_item(
            resp.json['executions'],
            workflow_name='main_wf'
        )
        self.assertIsNone(main_wf.get('output'))



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import datetime
import eventlet
import random
import threading

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
from osprofiler import profiler

from mistral import context
from mistral.db import utils as db_utils
from mistral.db.v2 import api as db_api
from mistral import exceptions as exc
from mistral.scheduler import base as sched_base
from mistral_lib import utils


LOG = logging.getLogger(__name__)


def _schedule_call(factory_method_path, target_method_name,
                   run_after, serializers=None, key=None, **method_args):
    """Schedules call and lately invokes target_method.

    Add this call specification to DB, and then after run_after
    seconds service CallScheduler invokes the target_method.

    :param factory_method_path: Full python-specific path to
        factory method that creates a target object that the call will be
        made against.
    :param target_method_name: Name of a method which will be invoked.
    :param run_after: Value in seconds.
    :param serializers: map of argument names and their serializer class
        paths. Use when an argument is an object of specific type, and needs
        to be serialized. Example:
        { "result": "mistral.utils.serializer.ResultSerializer"}
        Serializer for the object type must implement serializer interface
        in mistral/utils/serializer.py
    :param key: Key which can potentially be used for squashing similar
        delayed calls.
    :param method_args: Target method keyword arguments.
    """
    ctx_serializer = context.RpcContextSerializer()

    ctx = (
        ctx_serializer.serialize_context(context.ctx())
        if context.has_ctx() else {}
    )

    execution_time = (utils.utc_now_sec() +
                      datetime.timedelta(seconds=run_after))

    if serializers:
        for arg_name, serializer_path in serializers.items():
            if arg_name not in method_args:
                raise exc.MistralException(
                    "Serializable method argument %s"
                    " not found in method_args=%s"
                    % (arg_name, method_args))
            try:
                serializer = importutils.import_class(serializer_path)()
            except ImportError as e:
                raise ImportError(
                    "Cannot import class %s: %s" % (serializer_path, e)
                )

            method_args[arg_name] = serializer.serialize(method_args[arg_name])

    values = {
        'factory_method_path': factory_method_path,
        'target_method_name': target_method_name,
        'execution_time': execution_time,
        'auth_context': ctx,
        'serializers': serializers,
        'key': key,
        'method_arguments': method_args,
        'processing': False
    }

    db_api.create_delayed_call(values)


class LegacyScheduler(sched_base.Scheduler):
    def __init__(self, conf):
        self._stopped = False
        self._thread = threading.Thread(target=self._loop)
        self._thread.daemon = True
        self._fixed_delay = conf.fixed_delay
        self._random_delay = conf.random_delay
        self._batch_size = conf.batch_size

    def schedule(self, job, allow_redistribute=False):
        _schedule_call(
            job.target_factory_func_name,
            job.func_name,
            job.run_after,
            serializers=job.func_arg_serializers,
            key=job.key,
            **job.func_args
        )

    def has_scheduled_jobs(self, **filters):
        return db_api.get_delayed_calls_count(**filters) > 0

    def start(self):
        self._thread.start()

    def stop(self, graceful=False):
        self._stopped = True

        if graceful:
            self._thread.join()

    def _loop(self):
        # Scheduler runs jobs in an separate thread that's neither related
        # to an RPC nor a REST request processing thread. So we need to
        # initialize a profiler specifically for this thread.
        if cfg.CONF.profiler.enabled:
            profiler.init(cfg.CONF.profiler.hmac_keys)

        while not self._stopped:
            LOG.debug("Starting Scheduler loop [scheduler=%s]...", self)

            try:
                self._process_delayed_calls()
            except Exception:
                LOG.exception(
                    "Scheduler failed to process delayed calls"
                    " due to unexpected exception."
                )

            eventlet.sleep(
                self._fixed_delay +
                random.Random().randint(0, self._random_delay * 1000) * 0.001
            )

    def _process_delayed_calls(self, ctx=None):
        """Run delayed required calls.

        This algorithm should work with transactions having at least
        'READ-COMMITTED' isolation mode.

        :param ctx: Auth context.
        """

        # Select and capture calls matching time criteria.
        db_calls = self._capture_calls(self._batch_size)

        if not db_calls:
            return

        # Determine target methods, deserialize arguments etc.
        prepared_calls = self._prepare_calls(db_calls)

        # Invoke prepared calls.
        self._invoke_calls(prepared_calls)

        # Delete invoked calls from DB.
        self.delete_calls(db_calls)

    @staticmethod
    @db_utils.retry_on_db_error
    def _capture_calls(batch_size):
        """Captures delayed calls eligible for processing (based on time).

        The intention of this method is to select delayed calls based on time
        criteria and mark them in DB as being processed so that no other
        threads could process them in parallel.

        :return: A list of delayed calls captured for further processing.
        """
        result = []

        time_filter = utils.utc_now_sec() + datetime.timedelta(seconds=1)

        with db_api.transaction():
            candidates = db_api.get_delayed_calls_to_start(
                time_filter,
                batch_size
            )

            for call in candidates:
                # Mark this delayed call has been processed in order to
                # prevent calling from parallel transaction.
                db_call, updated_cnt = db_api.update_delayed_call(
                    id=call.id,
                    values={'processing': True},
                    query_filter={'processing': False}
                )

                # If updated_cnt != 1 then another scheduler
                # has already updated it.
                if updated_cnt == 1:
                    result.append(db_call)

        LOG.debug("Scheduler captured %s delayed calls.", len(result))

        return result

    @staticmethod
    def _prepare_calls(raw_calls):
        """Prepares delayed calls for invocation.

        After delayed calls were selected from DB they still need to be
        prepared for further usage, we need to build final target methods
        and deserialize arguments, if needed.

        :param raw_calls: Delayed calls fetched from DB (DB models).
        :return: A list of tuples (target_auth_context, target_method,
         method_args) where all data is properly deserialized.
        """

        result = []

        for call in raw_calls:
            LOG.debug(
                'Preparing next delayed call. '
                '[ID=%s, factory_method_path=%s, target_method_name=%s, '
                'method_arguments=%s]', call.id, call.factory_method_path,
                call.target_method_name, call.method_arguments
            )

            target_auth_context = copy.deepcopy(call.auth_context)

            if call.factory_method_path:
                factory = importutils.import_class(call.factory_method_path)

                target_method = getattr(factory(), call.target_method_name)
            else:
                target_method = importutils.import_class(
                    call.target_method_name
                )

            method_args = copy.deepcopy(call.method_arguments)

            if call.serializers:
                # Deserialize arguments.
                for arg_name, ser_path in call.serializers.items():
                    serializer = importutils.import_class(ser_path)()

                    deserialized = serializer.deserialize(
                        method_args[arg_name]
                    )

                    method_args[arg_name] = deserialized

            result.append((target_auth_context, target_method, method_args))

        return result

    @staticmethod
    def _invoke_calls(delayed_calls):
        """Invokes prepared delayed calls.

        :param delayed_calls: Prepared delayed calls represented as tuples
        (target_auth_context, target_method, method_args).
        """

        ctx_serializer = context.RpcContextSerializer()

        for (target_auth_context, target_method, method_args) in delayed_calls:
            try:
                # Set the correct context for the method.
                ctx_serializer.deserialize_context(target_auth_context)

                # Invoke the method.
                target_method(**method_args)
            except Exception as e:
                LOG.exception(
                    "Delayed call failed, method: %s, exception: %s",
                    target_method,
                    e
                )
            finally:
                # Remove context.
                context.set_ctx(None)

    @staticmethod
    @db_utils.retry_on_db_error
    def delete_calls(db_calls):
        """Deletes delayed calls.

        :param db_calls: Delayed calls to delete from DB.
        """
        try:
            db_api.delete_delayed_calls(id={'in': [c.id for c in db_calls]})
        except Exception as e:
            LOG.error(
                "Failed to delete all delayed calls [exception=%s]", e
            )

            # We have to re-raise any exception because the transaction
            # would be already invalid anyway. If it's a deadlock then
            # it will be handled.
            raise e

        LOG.debug("Scheduler deleted %s delayed calls.", len(db_calls))



# Copyright 2014 - Mirantis, Inc.
# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg

from mistral.db.v2 import api as db_api
from mistral.lang import parser as spec_parser
from mistral.services import workbooks as wb_service
from mistral.tests.unit import base

# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')

WORKBOOK = """
---
version: '2.0'

name: my_wb
tags: [test]

actions:
  concat:
    base: std.echo
    base-input:
      output: "{$.str1}{$.str2}"

workflows:
  wf1:
#Sample Comment 1
    type: reverse
    tags: [wf_test]
    input:
      - param1
    output:
      result: "{$.result}"

    tasks:
      task1:
        action: std.echo output="{$.param1}"
        publish:
          result: "{$}"

  wf2:
    type: direct
    output:
      result: "{$.result}"

    tasks:
      task1:
        workflow: my_wb.wf1 param1='Hi' task_name='task1'
        publish:
          result: "The result of subworkflow is '{$.final_result}'"
"""

WORKBOOK_WF1_DEFINITION = """wf1:
#Sample Comment 1
  type: reverse
  tags: [wf_test]
  input:
    - param1
  output:
    result: "{$.result}"

  tasks:
    task1:
      action: std.echo output="{$.param1}"
      publish:
        result: "{$}"
"""

WORKBOOK_WF2_DEFINITION = """wf2:
  type: direct
  output:
    result: "{$.result}"

  tasks:
    task1:
      workflow: my_wb.wf1 param1='Hi' task_name='task1'
      publish:
        result: "The result of subworkflow is '{$.final_result}'"
"""

UPDATED_WORKBOOK = """
---
version: '2.0'

name: my_wb
tags: [test]

actions:
  concat:
    base: std.echo
    base-input:
      output: "{$.str1}{$.str2}"

workflows:
  wf1:
    type: direct
    output:
      result: "{$.result}"

    tasks:
      task1:
        workflow: my_wb.wf2 param1='Hi' task_name='task1'
        publish:
          result: "The result of subworkflow is '{$.final_result}'"

  wf2:
    type: reverse
    input:
      - param1
    output:
      result: "{$.result}"

    tasks:
      task1:
        action: std.echo output="{$.param1}"
        publish:
          result: "{$}"
"""

UPDATED_WORKBOOK_WF1_DEFINITION = """wf1:
  type: direct
  output:
    result: "{$.result}"

  tasks:
    task1:
      workflow: my_wb.wf2 param1='Hi' task_name='task1'
      publish:
        result: "The result of subworkflow is '{$.final_result}'"
"""

UPDATED_WORKBOOK_WF2_DEFINITION = """wf2:
  type: reverse
  input:
    - param1
  output:
    result: "{$.result}"

  tasks:
    task1:
      action: std.echo output="{$.param1}"
      publish:
        result: "{$}"
"""

ACTION_DEFINITION = """concat:
  base: std.echo
  base-input:
    output: "{$.str1}{$.str2}"
"""


class WorkbookServiceTest(base.DbTestCase):
    def test_create_workbook(self):
        namespace = 'test_workbook_service_0123_namespace'

        wb_db = wb_service.create_workbook_v2(WORKBOOK, namespace=namespace)

        self.assertIsNotNone(wb_db)
        self.assertEqual('my_wb', wb_db.name)
        self.assertEqual(namespace, wb_db.namespace)
        self.assertEqual(WORKBOOK, wb_db.definition)
        self.assertIsNotNone(wb_db.spec)
        self.assertListEqual(['test'], wb_db.tags)

        db_actions = db_api.get_action_definitions(
            name='my_wb.concat',
            namespace=namespace
        )

        self.assertEqual(1, len(db_actions))

        # Action.
        action_db = self._assert_single_item(db_actions, name='my_wb.concat')

        self.assertFalse(action_db.is_system)

        action_spec = spec_parser.get_action_spec(action_db.spec)

        self.assertEqual('concat', action_spec.get_name())
        self.assertEqual('std.echo', action_spec.get_base())
        self.assertEqual(ACTION_DEFINITION, action_db.definition)

        db_wfs = db_api.get_workflow_definitions()

        self.assertEqual(2, len(db_wfs))

        # Workflow 1.
        wf1_db = self._assert_single_item(db_wfs, name='my_wb.wf1')
        wf1_spec = spec_parser.get_workflow_spec(wf1_db.spec)

        self.assertEqual('wf1', wf1_spec.get_name())
        self.assertEqual('reverse', wf1_spec.get_type())
        self.assertListEqual(['wf_test'], wf1_spec.get_tags())
        self.assertListEqual(['wf_test'], wf1_db.tags)
        self.assertEqual(namespace, wf1_db.namespace)
        self.assertEqual(WORKBOOK_WF1_DEFINITION, wf1_db.definition)

        # Workflow 2.
        wf2_db = self._assert_single_item(db_wfs, name='my_wb.wf2')
        wf2_spec = spec_parser.get_workflow_spec(wf2_db.spec)

        self.assertEqual('wf2', wf2_spec.get_name())
        self.assertEqual('direct', wf2_spec.get_type())
        self.assertEqual(namespace, wf2_db.namespace)
        self.assertEqual(WORKBOOK_WF2_DEFINITION, wf2_db.definition)

    def test_create_same_workbook_in_different_namespaces(self):
        first_namespace = 'first_namespace'
        second_namespace = 'second_namespace'

        first_wb = wb_service.create_workbook_v2(WORKBOOK,
                                                 namespace=first_namespace)
        self.assertIsNotNone(first_wb)
        self.assertEqual('my_wb', first_wb.name)
        self.assertEqual(first_namespace, first_wb.namespace)

        second_wb = wb_service.create_workbook_v2(WORKBOOK,
                                                  namespace=second_namespace)
        self.assertIsNotNone(second_wb)
        self.assertEqual('my_wb', second_wb.name)
        self.assertEqual(second_namespace, second_wb.namespace)

    def test_create_workbook_with_default_namespace(self):
        wb_db = wb_service.create_workbook_v2(WORKBOOK)

        self.assertIsNotNone(wb_db)
        self.assertEqual('my_wb', wb_db.name)
        self.assertEqual('', wb_db.namespace)

        db_api.delete_workbook('my_wb')

    def test_update_workbook(self):
        namespace = 'test_workbook_service_0123_namespace'

        # Create workbook.
        wb_db = wb_service.create_workbook_v2(WORKBOOK, namespace=namespace)

        self.assertIsNotNone(wb_db)
        self.assertEqual(2, len(db_api.get_workflow_definitions()))

        # Update workbook.
        wb_db = wb_service.update_workbook_v2(
            UPDATED_WORKBOOK,
            namespace=namespace
        )

        self.assertIsNotNone(wb_db)
        self.assertEqual('my_wb', wb_db.name)
        self.assertEqual(namespace, wb_db.namespace)
        self.assertEqual(UPDATED_WORKBOOK, wb_db.definition)
        self.assertListEqual(['test'], wb_db.tags)

        db_wfs = db_api.get_workflow_definitions()

        self.assertEqual(2, len(db_wfs))

        # Workflow 1.
        wf1_db = self._assert_single_item(db_wfs, name='my_wb.wf1')
        wf1_spec = spec_parser.get_workflow_spec(wf1_db.spec)

        self.assertEqual('wf1', wf1_spec.get_name())
        self.assertEqual('direct', wf1_spec.get_type())
        self.assertEqual(namespace, wf1_db.namespace)
        self.assertEqual(UPDATED_WORKBOOK_WF1_DEFINITION, wf1_db.definition)

        # Workflow 2.
        wf2_db = self._assert_single_item(db_wfs, name='my_wb.wf2')
        wf2_spec = spec_parser.get_workflow_spec(wf2_db.spec)

        self.assertEqual('wf2', wf2_spec.get_name())
        self.assertEqual('reverse', wf2_spec.get_type())
        self.assertEqual(namespace, wf2_db.namespace)
        self.assertEqual(UPDATED_WORKBOOK_WF2_DEFINITION, wf2_db.definition)

    def test_delete_workbook(self):
        namespace = 'pqr'

        # Create workbook.
        wb_service.create_workbook_v2(WORKBOOK, namespace=namespace)

        db_wfs = db_api.get_workflow_definitions()
        db_actions = db_api.get_action_definitions(name='my_wb.concat',
                                                   namespace=namespace)

        self.assertEqual(2, len(db_wfs))
        self.assertEqual(1, len(db_actions))

        db_api.delete_workbook('my_wb', namespace=namespace)

        db_wfs = db_api.get_workflow_definitions()
        db_actions = db_api.get_action_definitions(name='my_wb.concat',
                                                   namespace=namespace)
        # Deleting workbook shouldn't delete workflows and actions
        self.assertEqual(2, len(db_wfs))
        self.assertEqual(1, len(db_actions))



#    Copyright 2019 OpenStack Foundation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


from mistral import exceptions as exc
from mistral.services import workflows as wf_service
from mistral.tests.unit.engine import base
from testtools import ExpectedException


# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


class NameValidationTest(base.EngineTestCase):
    @staticmethod
    def test_workflow_name_validation():
        wf = """
        version: 2.0
        wf name with space:
          tasks:
            t1:
              action: a1
        """
        with ExpectedException(exc.InvalidModelException,
                               "Name 'wf name with space' "
                               "must not contain spaces"):
            wf_service.create_workflows(wf)



# Copyright 2019 - Nokia Networks, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.services import workbooks as wb_service
from mistral.services import workflows as wf_service
from mistral.tests.unit.api import base
from mistral.tests.unit.engine import base as engine_base
from mistral.workflow import states


class TestExecutionReportController(base.APITest, engine_base.EngineTestCase):
    def test_simple_sequence_wf(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.noop
              on-success: task2

            task2:
              action: std.fail
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_error(wf_ex.id)

        resp = self.app.get('/v2/executions/%s/report' % wf_ex.id)

        self.assertEqual(200, resp.status_int)

        # Now let's verify the response structure

        self.assertIn('root_workflow_execution', resp.json)

        root_wf_ex = resp.json['root_workflow_execution']

        self.assertIsInstance(root_wf_ex, dict)
        self.assertEqual(wf_ex.id, root_wf_ex['id'])
        self.assertEqual(wf_ex.name, root_wf_ex['name'])
        self.assertEqual(states.ERROR, root_wf_ex['state'])
        self.assertGreater(len(root_wf_ex['state_info']), 0)

        tasks = root_wf_ex['task_executions']

        self.assertIsInstance(tasks, list)

        self.assertEqual(2, len(tasks))

        # Verify task1 info.
        task1 = self._assert_single_item(
            tasks,
            name='task1',
            state=states.SUCCESS
        )

        self.assertEqual(0, len(task1['workflow_executions']))
        self.assertEqual(1, len(task1['action_executions']))

        task1_action = task1['action_executions'][0]

        self.assertEqual(states.SUCCESS, task1_action['state'])
        self.assertEqual('std.noop', task1_action['name'])

        # Verify task2 info.
        task2 = self._assert_single_item(
            tasks,
            name='task2',
            state=states.ERROR
        )

        self.assertEqual(1, len(task2['action_executions']))

        task2_action = task2['action_executions'][0]

        self.assertEqual(0, len(task2['workflow_executions']))
        self.assertEqual(states.ERROR, task2_action['state'])

        # Verify statistics.
        stat = resp.json['statistics']

        self.assertEqual(1, stat['error_tasks_count'])
        self.assertEqual(0, stat['idle_tasks_count'])
        self.assertEqual(0, stat['paused_tasks_count'])
        self.assertEqual(0, stat['running_tasks_count'])
        self.assertEqual(1, stat['success_tasks_count'])
        self.assertEqual(2, stat['total_tasks_count'])

    def test_nested_wf(self):
        wb_text = """---
        version: '2.0'

        name: wb

        workflows:
          parent_wf:
            tasks:
              task1:
                action: std.noop
                on-success: task2

              task2:
                workflow: sub_wf
                on-success: task3

              task3:
                action: std.fail

          sub_wf:
            tasks:
              task1:
                action: std.noop
                on-success: task2

              task2:
                action: std.fail
        """

        wb_service.create_workbook_v2(wb_text)

        wf_ex = self.engine.start_workflow('wb.parent_wf')

        self.await_workflow_error(wf_ex.id)

        resp = self.app.get('/v2/executions/%s/report' % wf_ex.id)

        self.assertEqual(200, resp.status_int)

        # Now let's verify the response structure

        self.assertIn('root_workflow_execution', resp.json)

        root_wf_ex = resp.json['root_workflow_execution']

        self.assertIsInstance(root_wf_ex, dict)
        self.assertEqual('wb.parent_wf', root_wf_ex['name'])
        self.assertEqual(states.ERROR, root_wf_ex['state'])
        self.assertGreater(len(root_wf_ex['state_info']), 0)

        tasks = root_wf_ex['task_executions']

        self.assertIsInstance(tasks, list)

        self.assertEqual(2, len(tasks))

        # Verify task1 info.
        task1 = self._assert_single_item(tasks, name='task1')

        self.assertEqual(states.SUCCESS, task1['state'])
        self.assertEqual(0, len(task1['workflow_executions']))
        self.assertEqual(1, len(task1['action_executions']))

        task1_action = task1['action_executions'][0]
        self.assertEqual(states.SUCCESS, task1_action['state'])
        self.assertEqual('std.noop', task1_action['name'])

        # Verify task2 info.
        task2 = self._assert_single_item(tasks, name='task2')

        self.assertEqual(states.ERROR, task2['state'])
        self.assertEqual(0, len(task2['action_executions']))
        self.assertEqual(1, len(task2['workflow_executions']))

        sub_wf_entry = task2['workflow_executions'][0]
        self.assertEqual(states.ERROR, sub_wf_entry['state'])

        sub_wf_tasks = sub_wf_entry['task_executions']

        self.assertEqual(2, len(sub_wf_tasks))

        sub_wf_task1 = self._assert_single_item(
            sub_wf_tasks,
            name='task1',
            state=states.SUCCESS
        )
        sub_wf_task2 = self._assert_single_item(
            sub_wf_tasks,
            name='task2',
            state=states.ERROR
        )

        self.assertEqual(1, len(sub_wf_task1['action_executions']))
        self.assertEqual(
            states.SUCCESS,
            sub_wf_task1['action_executions'][0]['state']
        )

        self.assertEqual(1, len(sub_wf_task2['action_executions']))
        self.assertEqual(
            states.ERROR,
            sub_wf_task2['action_executions'][0]['state']
        )

        # Verify statistics.
        stat = resp.json['statistics']

        self.assertEqual(2, stat['error_tasks_count'])
        self.assertEqual(0, stat['idle_tasks_count'])
        self.assertEqual(0, stat['paused_tasks_count'])
        self.assertEqual(0, stat['running_tasks_count'])
        self.assertEqual(2, stat['success_tasks_count'])
        self.assertEqual(4, stat['total_tasks_count'])

    def test_statistics_only(self):
        wb_text = """---
        version: '2.0'

        name: wb

        workflows:
          parent_wf:
            tasks:
              task1:
                action: std.noop
                on-success: task2

              task2:
                workflow: sub_wf
                on-success: task3

              task3:
                action: std.fail

          sub_wf:
            tasks:
              task1:
                action: std.noop
                on-success: task2

              task2:
                action: std.fail
        """

        wb_service.create_workbook_v2(wb_text)

        wf_ex = self.engine.start_workflow('wb.parent_wf')

        self.await_workflow_error(wf_ex.id)

        resp = self.app.get(
            '/v2/executions/%s/report?statistics_only=True' % wf_ex.id
        )

        self.assertEqual(200, resp.status_int)

        stat = resp.json['statistics']

        self.assertEqual(2, stat['error_tasks_count'])
        self.assertEqual(0, stat['idle_tasks_count'])
        self.assertEqual(0, stat['paused_tasks_count'])
        self.assertEqual(0, stat['running_tasks_count'])
        self.assertEqual(2, stat['success_tasks_count'])
        self.assertEqual(4, stat['total_tasks_count'])

        self.assertNotIn('root_workflow_execution', resp.json)

    def test_estimated_time(self):
        wf_text = """---
        version: '2.0'
        wf1:
          tasks:
            task1:
              action: std.sleep seconds=2
        """

        wf_service.create_workflows(wf_text)

        for i in range(3):
            wf_ex = self.engine.start_workflow('wf1')

        self.await_workflow_success(wf_ex.id)

        wf_ex = self.engine.start_workflow('wf1')

        resp = self.app.get(
            '/v2/executions/%s/report?statistics_only=True' % wf_ex.id)

        self.assertEqual(200, resp.status_int)

        stat = resp.json['statistics']
        self.assertTrue(stat['estimated_time'] > 0)

    def test_estimated_time_no_previous_executions(self):
        wf_text = """---
        version: '2.0'
        wf1:
          tasks:
            task1:
              action: std.sleep seconds=5
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf1')

        resp = self.app.get(
            '/v2/executions/%s/report?statistics_only=True' % wf_ex.id)

        self.assertEqual(200, resp.status_int)

        stat = resp.json['statistics']

        self.assertTrue(stat['estimated_time'] == -1)

    def test_nested_wf_errors_only(self):
        wb_text = """---
        version: '2.0'

        name: wb

        workflows:
          parent_wf:
            tasks:
              task1:
                action: std.noop
                on-success: task2

              task2:
                workflow: sub_wf
                on-success: task3

              task3:
                action: std.fail

          sub_wf:
            tasks:
              task1:
                action: std.noop
                on-success: task2

              task2:
                action: std.fail
        """

        wb_service.create_workbook_v2(wb_text)

        wf_ex = self.engine.start_workflow('wb.parent_wf')

        self.await_workflow_error(wf_ex.id)

        resp = self.app.get(
            '/v2/executions/%s/report?errors_only=true' % wf_ex.id
        )

        self.assertEqual(200, resp.status_int)

        # Now let's verify the response structure

        self.assertIn('root_workflow_execution', resp.json)

        root_wf_ex = resp.json['root_workflow_execution']

        self.assertIsInstance(root_wf_ex, dict)
        self.assertEqual('wb.parent_wf', root_wf_ex['name'])
        self.assertEqual(states.ERROR, root_wf_ex['state'])
        self.assertGreater(len(root_wf_ex['state_info']), 0)

        tasks = root_wf_ex['task_executions']

        self.assertIsInstance(tasks, list)

        self.assertEqual(1, len(tasks))

        # There must be only task2 in the response.
        # Verify task2 info.
        task2 = self._assert_single_item(tasks, name='task2')

        self.assertEqual(states.ERROR, task2['state'])
        self.assertEqual(0, len(task2['action_executions']))
        self.assertEqual(1, len(task2['workflow_executions']))

        sub_wf_entry = task2['workflow_executions'][0]
        self.assertEqual(states.ERROR, sub_wf_entry['state'])

        sub_wf_tasks = sub_wf_entry['task_executions']

        self.assertEqual(1, len(sub_wf_tasks))

        sub_wf_task2 = self._assert_single_item(
            sub_wf_tasks,
            name='task2',
            state=states.ERROR
        )

        self.assertEqual(1, len(sub_wf_task2['action_executions']))
        self.assertEqual(
            states.ERROR,
            sub_wf_task2['action_executions'][0]['state']
        )

        # Verify statistics.
        stat = resp.json['statistics']

        self.assertEqual(2, stat['error_tasks_count'])
        self.assertEqual(0, stat['idle_tasks_count'])
        self.assertEqual(0, stat['paused_tasks_count'])
        self.assertEqual(0, stat['running_tasks_count'])
        self.assertEqual(0, stat['success_tasks_count'])
        self.assertEqual(2, stat['total_tasks_count'])

    def test_nested_wf_max_depth(self):
        wb_text = """---
        version: '2.0'

        name: wb

        workflows:
          parent_wf:
            tasks:
              task1:
                action: std.noop
                on-success: task2

              task2:
                workflow: sub_wf
                on-success: task3

              task3:
                action: std.fail

          sub_wf:
            tasks:
              task1:
                action: std.noop
                on-success: task2

              task2:
                action: std.fail
        """

        wb_service.create_workbook_v2(wb_text)

        wf_ex = self.engine.start_workflow('wb.parent_wf')

        self.await_workflow_error(wf_ex.id)

        resp = self.app.get('/v2/executions/%s/report?max_depth=0' % wf_ex.id)

        self.assertEqual(200, resp.status_int)

        # Now let's verify the response structure

        self.assertIn('root_workflow_execution', resp.json)

        root_wf_ex = resp.json['root_workflow_execution']

        self.assertIsInstance(root_wf_ex, dict)
        self.assertEqual('wb.parent_wf', root_wf_ex['name'])
        self.assertEqual(states.ERROR, root_wf_ex['state'])
        self.assertGreater(len(root_wf_ex['state_info']), 0)

        tasks = root_wf_ex['task_executions']

        self.assertIsInstance(tasks, list)

        self.assertEqual(2, len(tasks))

        # Verify task1 info.
        task1 = self._assert_single_item(tasks, name='task1')

        self.assertEqual(states.SUCCESS, task1['state'])
        self.assertEqual(0, len(task1['workflow_executions']))
        self.assertEqual(1, len(task1['action_executions']))

        task1_action = task1['action_executions'][0]
        self.assertEqual(states.SUCCESS, task1_action['state'])
        self.assertEqual('std.noop', task1_action['name'])

        # Verify task2 info.
        task2 = self._assert_single_item(tasks, name='task2')

        self.assertEqual(states.ERROR, task2['state'])
        self.assertEqual(0, len(task2['action_executions']))
        self.assertEqual(1, len(task2['workflow_executions']))

        sub_wf_entry = task2['workflow_executions'][0]
        self.assertEqual(states.ERROR, sub_wf_entry['state'])

        # We still must have an entry for the subworkflow itself
        # but it must not have info about task executions because
        # we've now limited max depth.
        self.assertNotIn('task_executions', sub_wf_entry)

        # Verify statistics.
        stat = resp.json['statistics']

        self.assertEqual(1, stat['error_tasks_count'])
        self.assertEqual(0, stat['idle_tasks_count'])
        self.assertEqual(0, stat['paused_tasks_count'])
        self.assertEqual(0, stat['running_tasks_count'])
        self.assertEqual(1, stat['success_tasks_count'])
        self.assertEqual(2, stat['total_tasks_count'])

    def test_retry_count(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.noop
              on-success: task2

            task2:
              action: std.fail
              retry: delay=0.1 count=3
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_error(wf_ex.id)

        resp = self.app.get('/v2/executions/%s/report' % wf_ex.id)

        self.assertEqual(200, resp.status_int)

        # Now let's verify the response structure

        root_wf_ex = resp.json['root_workflow_execution']

        tasks = root_wf_ex['task_executions']

        self.assertEqual(2, len(tasks))

        # Verify task1 presence.
        self._assert_single_item(tasks, name='task1', state=states.SUCCESS)

        # Verify task2 info.
        task2 = self._assert_single_item(
            tasks,
            name='task2',
            state=states.ERROR
        )

        self.assertEqual(3, task2['retry_count'])



# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import collections

from oslo_config import cfg
from oslo_log import log as logging


from mistral_lib import actions as ml_actions
from mistral_lib.utils import inspect_utils as i_utils


CONF = cfg.CONF

LOG = logging.getLogger(__name__)


def _build_action_descriptor(name, action_cls):
    action_cls_attrs = i_utils.get_public_fields(action_cls)

    return ml_actions.PythonActionDescriptor(
        name,
        action_cls,
        action_cls_attrs
    )


class TestActionProvider(ml_actions.ActionProvider):
    """Action provider for tests.

    It allows to register python actions with a direct call.
    """

    def __init__(self, name='test'):
        super().__init__(name)

        self._action_descs = collections.OrderedDict()

    def register_python_action(self, action_name, action_cls):
        self._action_descs[action_name] = _build_action_descriptor(
            action_name,
            action_cls
        )

    def cleanup(self):
        self._action_descs.clear()

    def find(self, action_name, namespace=None):
        return self._action_descs.get(action_name)

    def find_all(self, namespace=None, limit=None, sort_fields=None,
                 sort_dirs=None, **filters):
        # TODO(rakhmerov): Apply sort_keys, sort_dirs, and filters.
        return self._action_descs.values()



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import cachetools
from unittest import mock

from oslo_config import cfg

from mistral.db.v2 import api as db_api
from mistral.rpc import clients as rpc_clients
from mistral.services import workflows as wf_service
from mistral.tests.unit.engine import base
from mistral.workflow import states


# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


class ActionHeartbeatCheckerTest(base.EngineTestCase):
    def setUp(self):
        # We need to override configuration values before starting engine.
        self.override_config('check_interval', 1, 'action_heartbeat')
        self.override_config('max_missed_heartbeats', 1, 'action_heartbeat')
        self.override_config('first_heartbeat_timeout', 0, 'action_heartbeat')

        super(ActionHeartbeatCheckerTest, self).setUp()

        self.override_config('type', 'remote', 'executor')

    # Make sure actions are not sent to an executor.
    @mock.patch.object(
        rpc_clients.ExecutorClient,
        'run_action',
        mock.MagicMock()
    )
    def test_fail_action_with_missing_heartbeats(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.noop
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        # The workflow should fail because the action of "task1" should be
        # failed automatically by the action execution heartbeat checker.
        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            t_execs = wf_ex.task_executions

            t_ex = self._assert_single_item(
                t_execs,
                name='task1',
                state=states.ERROR
            )

            a_execs = db_api.get_action_executions(task_execution_id=t_ex.id)

            self._assert_single_item(
                a_execs,
                name='std.noop',
                state=states.ERROR
            )

    # Make sure actions are not sent to an executor.
    @mock.patch.object(
        rpc_clients.ExecutorClient,
        'run_action',
        mock.MagicMock()
    )
    @mock.patch.object(
        cachetools.LRUCache,
        '__getitem__',
        mock.MagicMock(side_effect=KeyError)
    )
    @mock.patch.object(
        cachetools.LRUCache,
        '__contains__',
        mock.MagicMock(return_value=False)
    )
    def test_fail_action_with_missing_heartbeats_wf_spec_not_cached(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.noop
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        # The workflow should fail because the action of "task1" should be
        # failed automatically by the action execution heartbeat checker.
        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            t_execs = wf_ex.task_executions

            t_ex = self._assert_single_item(
                t_execs,
                name='task1',
                state=states.ERROR
            )

            a_execs = db_api.get_action_executions(task_execution_id=t_ex.id)

            self._assert_single_item(
                a_execs,
                name='std.noop',
                state=states.ERROR
            )



# Copyright 2016 NEC Corporation. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_concurrency import processutils
from oslo_config import cfg

from mistral.api import service
from mistral.tests.unit import base


class TestWSGIService(base.BaseTest):
    def setUp(self):
        super(TestWSGIService, self).setUp()

        self.override_config('enabled', False, group='cron_trigger')

    @mock.patch.object(service.wsgi, 'Server')
    def test_workers_set_default(self, wsgi_server):
        service_name = "mistral_api"

        with mock.patch('mistral.api.app.setup_app'):
            test_service = service.WSGIService(service_name)

            wsgi_server.assert_called_once_with(
                cfg.CONF,
                service_name,
                test_service.app,
                host='0.0.0.0',
                port=8989,
                use_ssl=False
            )

    def test_workers_set_correct_setting(self):
        self.override_config('api_workers', 8, group='api')

        with mock.patch('mistral.api.app.setup_app'):
            test_service = service.WSGIService("mistral_api")

            self.assertEqual(8, test_service.workers)

    def test_workers_set_zero_setting(self):
        self.override_config('api_workers', 0, group='api')

        with mock.patch('mistral.api.app.setup_app'):
            test_service = service.WSGIService("mistral_api")

            self.assertEqual(
                processutils.get_worker_count(),
                test_service.workers
            )

    @mock.patch.object(service.wsgi, 'Server')
    def test_wsgi_service_with_ssl_enabled(self, wsgi_server):
        self.override_config('enable_ssl_api', True, group='api')

        service_name = 'mistral_api'

        with mock.patch('mistral.api.app.setup_app'):
            srv = service.WSGIService(service_name)

            wsgi_server.assert_called_once_with(
                cfg.CONF,
                service_name,
                srv.app,
                host='0.0.0.0',
                port=8989,
                use_ssl=True
            )



# Copyright 2014 - Mirantis, Inc.
# Copyright 2014 - StackStorm, Inc.
# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging

from mistral.db.v2 import api as db_api
from mistral.services import adhoc_actions
from mistral_lib import utils


# TODO(rakhmerov): This module won't be needed after we add action providers

LOG = logging.getLogger(__name__)

ACTIONS_PATH = 'resources/actions'


def _register_preinstalled_adhoc_actions():
    action_paths = utils.get_file_list(ACTIONS_PATH)

    for action_path in action_paths:
        action_definition = open(action_path).read()

        adhoc_actions.create_or_update_actions(
            action_definition,
            scope='public'
        )


def sync_db():
    with db_api.transaction():
        _register_preinstalled_adhoc_actions()



# Copyright 2014 - Mirantis, Inc.
# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg

from mistral_lib import actions as ml_actions

from mistral.db.v2 import api as db_api
from mistral import exceptions as exc
from mistral.services import workbooks as wb_service
from mistral.tests.unit.engine import base
from mistral.workflow import states


# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')

WB = """
---
version: '2.0'

name: my_wb

actions:
  concat_twice:
    base: std.echo
    base-input:
      output: "<% $.s1 %>+<% $.s2 %>"
    input:
      - s1: "a"
      - s2
    output: "<% $ %> and <% $ %>"

  test_env:
    base: std.echo
    base-input:
      output: '{{ env().foo }}'

  nested_concat:
    base: my_wb.concat_twice
    base-input:
      s2: '{{ _.n2 }}'
    input:
      - n2: 'b'
    output:
      nested_concat: '{{ _ }}'

  missing_base:
    base: wrong
    input:
      - some_input

  nested_missing_base:
    base: missing_base
    input:
      - some_input

workflows:
  wf1:
    type: direct
    input:
      - str1
      - str2
    output:
      workflow_result: <% $.result %> # Access to execution context variables
      concat_task_result: <% task(concat).result %> # Same but via task name

    tasks:
      concat:
        action: concat_twice s1=<% $.str1 %> s2=<% $.str2 %>
        publish:
          result: <% task(concat).result %>

  wf2:
    type: direct
    input:
      - str1
      - str2
    output:
      workflow_result: <% $.result %> # Access to execution context variables
      concat_task_result: <% task(concat).result %> # Same but via task name

    tasks:
      concat:
        action: concat_twice s2=<% $.str2 %>
        publish:
          result: <% task(concat).result %>

  wf3:
    type: direct
    input:
      - str1
      - str2

    tasks:
      concat:
        action: concat_twice

  wf4:
    type: direct
    input:
      - str1
    output:
        workflow_result: '{{ _.printenv_result }}'

    tasks:
      printenv:
        action: test_env
        publish:
          printenv_result: '{{ task().result }}'

  wf5:
    type: direct
    output:
      workflow_result: '{{ _.nested_result }}'
    tasks:
      nested_test:
        action: nested_concat
        publish:
          nested_result: '{{ task().result }}'

  wf6:
    type: direct
    output:
      workflow_result: '{{ _.missing_result }}'
    tasks:
      missing_action:
        action: missing_base
        on-complete:
          - next_action
      next_action:
        publish:
          missing_result: 'Finished'

  wf7:
    type: direct
    output:
      workflow_result: '{{ _.missing_result }}'
    tasks:
      nested_missing_action:
        action: nested_missing_base
        on-complete:
          - next_action
      next_action:
        publish:
          missing_result: 'Finished'
"""


class AdhocActionsTest(base.EngineTestCase):
    def setUp(self):
        super(AdhocActionsTest, self).setUp()

        wb_service.create_workbook_v2(WB)

    def test_run_workflow_with_adhoc_action(self):
        wf_ex = self.engine.start_workflow(
            'my_wb.wf1',
            wf_input={'str1': 'a', 'str2': 'b'}
        )

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertDictEqual(
                {
                    'workflow_result': 'a+b and a+b',
                    'concat_task_result': 'a+b and a+b'
                },
                wf_ex.output
            )

    def test_run_adhoc_action_without_input_value(self):
        wf_ex = self.engine.start_workflow(
            'my_wb.wf2',
            wf_input={'str1': 'a', 'str2': 'b'}
        )

        self.await_workflow_success(wf_ex.id)

        expected_output = {
            'workflow_result': 'a+b and a+b',
            'concat_task_result': 'a+b and a+b'
        }

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertDictEqual(expected_output, wf_ex.output)

    def test_run_adhoc_action_without_sufficient_input_value(self):
        wf_ex = self.engine.start_workflow(
            'my_wb.wf3',
            wf_input={'str1': 'a', 'str2': 'b'}
        )
        self.await_workflow_error(wf_ex.id)
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
        self.assertIn("Invalid input", wf_ex.state_info)
        self.assertEqual(states.ERROR, wf_ex.state)

    def test_run_adhoc_action_with_env(self):
        wf_ex = self.engine.start_workflow(
            'my_wb.wf4',
            wf_input={'str1': 'a'},
            env={'foo': 'bar'}
        )

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertDictEqual(
                {
                    'workflow_result': 'bar'
                },
                wf_ex.output
            )

    def test_run_nested_adhoc_with_output(self):
        wf_ex = self.engine.start_workflow('my_wb.wf5')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            self.assertDictEqual(
                {
                    'workflow_result': {'nested_concat': 'a+b and a+b'}
                },
                wf_ex.output
            )

    def test_missing_adhoc_action_definition(self):
        wf_ex = self.engine.start_workflow('my_wb.wf6')

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

            task1 = self._assert_single_item(tasks, name='missing_action')

        self.assertEqual(states.ERROR, task1.state)

    def test_nested_missing_adhoc_action_definition(self):
        wf_ex = self.engine.start_workflow('my_wb.wf7')

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

            task1 = self._assert_single_item(
                tasks,
                name='nested_missing_action'
            )

        self.assertEqual(states.ERROR, task1.state)

    def test_adhoc_async_action(self):
        wb_text = """---
        version: '2.0'

        name: my_wb1

        actions:
          my_action:
            input:
              - my_param

            base: std.async_noop
            output: (((<% $ %>)))

        workflows:
          my_wf:
            tasks:
              task1:
                action: my_action my_param="asdfasdf"
        """

        wb_service.create_workbook_v2(wb_text)

        wf_ex = self.engine.start_workflow('my_wb1.my_wf')

        self.await_workflow_running(wf_ex.id, timeout=4)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            wf_ex_id = wf_ex.id
            task_execs = wf_ex.task_executions
            task1_ex = task_execs[0]
        self.engine.start_task(task1_ex.id, True, False, None, False, False)
        action_execs = db_api.get_action_executions(
            task_execution_id=task1_ex.id
        )
        a_ex_id = action_execs[0].id

        self.engine.on_action_complete(a_ex_id, ml_actions.Result(data='Hi!'))

        self.await_action_success(a_ex_id)
        self.await_workflow_success(wf_ex_id)

        with db_api.transaction(read_only=True):
            a_ex = db_api.get_action_execution(a_ex_id)

            self.assertEqual('(((Hi!)))', a_ex.output.get('result'))

    def test_adhoc_action_definition_with_namespace(self):
        namespace1 = 'ad-hoc_test'
        namespace2 = 'ad-hoc_test2'

        wb_text = """---

        version: '2.0'

        name: my_wb1

        actions:
          test_env:
            base: std.echo
            base-input:
              output: '{{ env().foo }}' # TODO(rakhmerov): It won't work.

        workflows:
          wf:
            input:
              - str1
            output:
                workflow_result: '{{ _.printenv_result }}'

            tasks:
              printenv:
                action: test_env
                publish:
                  printenv_result: '{{ task().result }}'
        """

        wb_service.create_workbook_v2(wb_text, namespace=namespace1)
        wb_service.create_workbook_v2(wb_text, namespace=namespace2)

        with db_api.transaction():
            action_defs = db_api.get_action_definitions(
                name='my_wb1.test_env'
            )

            self.assertEqual(2, len(action_defs))

            action_defs = db_api.get_action_definitions(
                name='my_wb1.test_env',
                namespace=namespace1
            )

            self.assertEqual(1, len(action_defs))

            action_defs = db_api.get_action_definitions(
                name='my_wb1.test_env',
                namespace=namespace2
            )

            self.assertEqual(1, len(action_defs))

            self.assertRaises(
                exc.DBEntityNotFoundError,
                db_api.get_action_definition,
                name='my_wb1.test_env'
            )

    def test_adhoc_action_execution_with_namespace(self):
        namespace = 'ad-hoc_test'

        wb_service.create_workbook_v2(WB, namespace=namespace)

        wf_ex = self.engine.start_workflow(
            'my_wb.wf4',
            wf_input={'str1': 'a'},
            env={'foo': 'bar'},
            wf_namespace=namespace
        )

        self.await_workflow_success(wf_ex.id, timeout=5)

        with db_api.transaction():
            action_execs = db_api.get_action_executions(
                name='my_wb.test_env',
                workflow_namespace=namespace
            )

            self.assertEqual(1, len(action_execs))

            self.assertEqual(namespace, action_execs[0].workflow_namespace)

    def test_adhoc_action_runtime_context_name(self):
        wf_ex = self.engine.start_workflow(
            'my_wb.wf4',
            wf_input={'str1': 'a'},
            env={'foo': 'bar'}
        )

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            action_execs = db_api.get_action_executions(name='my_wb.test_env')

            self.assertEqual(1, len(action_execs))



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2019 - NetCracker Technology Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import json
import re

from mistral import exceptions as exc
from mistral import expressions
from mistral.lang import types
from mistral.lang.v2 import base
from mistral.lang.v2 import on_clause
from mistral.lang.v2 import policies
from mistral.lang.v2 import publish
from mistral.lang.v2 import retry_policy
from mistral.workflow import states
from mistral_lib import utils

# Task types.
ACTION_TASK_TYPE = 'ACTION'
WORKFLOW_TASK_TYPE = 'WORKFLOW'


_expr_ptrns = [expressions.patterns[name] for name in expressions.patterns]

WITH_ITEMS_PTRN = re.compile(
    r"\s*([\w\d_\-]+)\s*in\s*(\[.+\]|%s)" % '|'.join(_expr_ptrns)
)

MAX_LENGTH_TASK_NAME = 255
# Length of a join task name must be less than or equal to maximum
# of task_executions unique_key and named_locks name. Their
# maximum equals 255.
# https://dev.mysql.com/doc/refman/5.6/en/innodb-restrictions.html
# For example: "join-task-" + "workflow execution id" + "-" +
# "task join name" = 255
# "task join name" = 255 - 36 - 1 - 10 = MAX_LENGTH_TASK_NAME - 47
MAX_LENGTH_JOIN_TASK_NAME = MAX_LENGTH_TASK_NAME - 47


class TaskSpec(base.BaseSpec):
    # See http://json-schema.org
    _polymorphic_key = ('type', 'direct')

    _schema = {
        "type": "object",
        "properties": {
            "type": types.WORKFLOW_TYPE,
            "action": types.NONEMPTY_STRING,
            "workflow": types.NONEMPTY_STRING,
            "input": {
                "oneOf": [
                    types.NONEMPTY_DICT,
                    types.NONEMPTY_STRING
                ]
            },
            "with-items": {
                "oneOf": [
                    types.NONEMPTY_STRING,
                    types.UNIQUE_STRING_LIST
                ]
            },
            "publish": types.NONEMPTY_DICT,
            "publish-on-error": types.NONEMPTY_DICT,
            "publish-on-skip": types.NONEMPTY_DICT,
            "retry": retry_policy.RetrySpec.get_schema(),
            "wait-before": types.EXPRESSION_OR_POSITIVE_INTEGER,
            "wait-after": types.EXPRESSION_OR_POSITIVE_INTEGER,
            "timeout": types.EXPRESSION_OR_POSITIVE_INTEGER,
            "pause-before": types.EXPRESSION_OR_BOOLEAN,
            "concurrency": types.EXPRESSION_OR_POSITIVE_INTEGER,
            "fail-on": types.EXPRESSION_OR_BOOLEAN,
            "target": types.NONEMPTY_STRING,
            "keep-result": types.EXPRESSION_OR_BOOLEAN,
            "safe-rerun": types.EXPRESSION_OR_BOOLEAN
        },
        "additionalProperties": False,
        "anyOf": [
            {
                "not": {
                    "type": "object",
                    "required": ["action", "workflow"]
                },
            },
            {
                "oneOf": [
                    {
                        "type": "object",
                        "required": ["action"]
                    },
                    {
                        "type": "object",
                        "required": ["workflow"]
                    }
                ]
            }
        ]
    }

    def __init__(self, data, validate):
        super(TaskSpec, self).__init__(data, validate)

        self._name = data['name']
        self._description = data.get('description')
        self._action = data.get('action')
        self._workflow = data.get('workflow')
        self._tags = data.get('tags', [])
        self._input = data.get('input', {})
        self._with_items = self._get_with_items_as_dict()
        self._publish = data.get('publish', {})
        self._publish_on_error = data.get('publish-on-error', {})
        self._publish_on_skip = data.get('publish-on-skip', {})
        self._policies = self._group_spec(
            policies.PoliciesSpec,
            'retry',
            'wait-before',
            'wait-after',
            'timeout',
            'pause-before',
            'concurrency',
            'fail-on'
        )
        self._target = data.get('target')
        self._keep_result = data.get('keep-result', True)
        self._safe_rerun = data.get('safe-rerun')

        self._process_action_and_workflow()

    def validate_schema(self):
        super(TaskSpec, self).validate_schema()

        self._validate_name()

        action = self._data.get('action')
        workflow = self._data.get('workflow')

        # Validate YAQL expressions.
        if action or workflow:
            inline_params = self._parse_cmd_and_input(action or workflow)[1]
            self.validate_expr(inline_params)

        self.validate_expr(self._data.get('input', {}))
        self.validate_expr(self._data.get('publish', {}))
        self.validate_expr(self._data.get('publish-on-error', {}))
        self.validate_expr(self._data.get('publish-on-skip', {}))
        self.validate_expr(self._data.get('keep-result', {}))
        self.validate_expr(self._data.get('safe-rerun', {}))

    def _validate_name(self):
        task_name = self._data.get('name')

        if len(task_name) > MAX_LENGTH_TASK_NAME:
            raise exc.InvalidModelException(
                "The length of a '{0}' task name must not exceed {1}"
                " symbols".format(task_name, MAX_LENGTH_TASK_NAME))

    def _get_with_items_as_dict(self):
        raw = self._data.get('with-items', [])

        with_items = {}

        if isinstance(raw, str):
            raw = [raw]

        for item in raw:
            if not isinstance(item, str):
                raise exc.InvalidModelException(
                    "'with-items' elements should be strings: %s" % self._data
                )

            match = re.match(WITH_ITEMS_PTRN, item)

            if not match:
                raise exc.InvalidModelException(
                    "Wrong format of 'with-items' property. Please use "
                    "format 'var in {[some, list] | <%% $.array %%> }: "
                    "%s" % self._data
                )

            match_groups = match.groups()
            var_name = match_groups[0]
            array = match_groups[1]

            # Validate YAQL expression that may follow after "in" for the
            # with-items syntax "var in {[some, list] | <% $.array %> }".
            self.validate_expr(array)

            if array.startswith('['):
                try:
                    array = json.loads(array)
                except Exception as e:
                    msg = ("Invalid array in 'with-items' clause: "
                           "%s, error: %s" % (array, str(e)))
                    raise exc.InvalidModelException(msg)

            with_items[var_name] = array

        return with_items

    def _process_action_and_workflow(self):
        params = {}

        if self._action:
            self._action, params = self._parse_cmd_and_input(self._action)
        elif self._workflow:
            self._workflow, params = self._parse_cmd_and_input(
                self._workflow)
        else:
            self._action = 'std.noop'

        utils.merge_dicts(self._input, params)

    def get_name(self):
        return self._name

    def get_description(self):
        return self._description

    def get_action_name(self):
        return self._action if self._action else None

    def get_workflow_name(self):
        return self._workflow

    def get_tags(self):
        return self._tags

    def get_input(self):
        return self._input

    def get_with_items(self):
        return self._with_items

    def get_policies(self):
        return self._policies

    def get_target(self):
        return self._target

    def get_publish(self, state):
        spec = None

        if state == states.SUCCESS and self._publish:
            spec = publish.PublishSpec(
                {'branch': self._publish},
                validate=self._validate
            )
        elif state == states.ERROR and self._publish_on_error:
            spec = publish.PublishSpec(
                {'branch': self._publish_on_error},
                validate=self._validate
            )
        elif state == states.SKIPPED and self._publish_on_skip:
            spec = publish.PublishSpec(
                {'branch': self._publish_on_skip},
                validate=self._validate
            )
        return spec

    def get_keep_result(self):
        return self._keep_result

    def get_safe_rerun(self):
        return self._safe_rerun

    def get_type(self):
        return (WORKFLOW_TASK_TYPE if self._workflow
                else ACTION_TASK_TYPE)


class DirectWorkflowTaskSpec(TaskSpec):
    _polymorphic_value = 'direct'

    _direct_workflow_schema = {
        "type": "object",
        "properties": {
            "type": {"enum": [_polymorphic_value]},
            "join": {
                "oneOf": [
                    {"enum": ["all", "one"]},
                    types.POSITIVE_INTEGER
                ]
            },
            "on-complete": on_clause.OnClauseSpec.get_schema(),
            "on-success": on_clause.OnClauseSpec.get_schema(),
            "on-error": on_clause.OnClauseSpec.get_schema(),
            "on-skip": on_clause.OnClauseSpec.get_schema()
        }
    }

    _schema = utils.merge_dicts(
        copy.deepcopy(TaskSpec._schema),
        _direct_workflow_schema
    )

    def __init__(self, data, validate):
        super(DirectWorkflowTaskSpec, self).__init__(data, validate)

        self._join = data.get('join')

        on_spec_cls = on_clause.OnClauseSpec

        self._on_complete = self._spec_property('on-complete', on_spec_cls)
        self._on_success = self._spec_property('on-success', on_spec_cls)
        self._on_error = self._spec_property('on-error', on_spec_cls)
        self._on_skip = self._spec_property('on-skip', on_spec_cls)

    def validate_semantics(self):
        # Validate YAQL expressions.
        self._validate_transitions(self._on_complete)
        self._validate_transitions(self._on_success)
        self._validate_transitions(self._on_error)
        self._validate_transitions(self._on_skip)

        if self._join:
            join_task_name = self.get_name()

            if len(join_task_name) > MAX_LENGTH_JOIN_TASK_NAME:
                raise exc.InvalidModelException(
                    "The length of a '{0}' join task name must not exceed {1} "
                    "symbols".format(join_task_name, MAX_LENGTH_JOIN_TASK_NAME)
                )

    def _validate_transitions(self, on_clause_spec):
        val = on_clause_spec.get_next() if on_clause_spec else []

        if not val:
            return

        [self.validate_expr(t)
            for t in ([val] if isinstance(val, str) else val)]

    def get_publish(self, state):
        spec = super(DirectWorkflowTaskSpec, self).get_publish(state)

        if self._on_complete and self._on_complete.get_publish():
            if spec:
                spec.merge(self._on_complete.get_publish())
            else:
                spec = self._on_complete.get_publish()

        if state == states.SUCCESS:
            on_clause = self._on_success
        elif state == states.ERROR:
            on_clause = self._on_error
        elif state == states.SKIPPED:
            on_clause = self._on_skip

        if on_clause and on_clause.get_publish():
            if spec:
                on_clause.get_publish().merge(spec)

            return on_clause.get_publish()

        return spec

    def get_join(self):
        return self._join

    def get_on_complete(self):
        return self._on_complete

    def get_on_success(self):
        return self._on_success

    def get_on_error(self):
        return self._on_error

    def get_on_skip(self):
        return self._on_skip


class ReverseWorkflowTaskSpec(TaskSpec):
    _polymorphic_value = 'reverse'

    _reverse_workflow_schema = {
        "type": "object",
        "properties": {
            "type": {"enum": [_polymorphic_value]},
            "requires": {
                "oneOf": [types.NONEMPTY_STRING, types.UNIQUE_STRING_LIST]
            }
        }
    }

    _schema = utils.merge_dicts(
        copy.deepcopy(TaskSpec._schema),
        _reverse_workflow_schema
    )

    def __init__(self, data, validate):
        super(ReverseWorkflowTaskSpec, self).__init__(data, validate)

        self._requires = data.get('requires', [])

    def get_requires(self):
        if isinstance(self._requires, str):
            return [self._requires]

        return self._requires


class TaskSpecList(base.BaseSpecList):
    item_class = TaskSpec



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keystoneauth1 import loading
from keystoneclient.v3 import client as ks_client
from oslo_config import cfg

from mistral import context

CONF = cfg.CONF


def client():
    ctx = context.ctx()
    auth_url = ctx.auth_uri or CONF.keystone_authtoken.www_authenticate_uri

    cl = ks_client.Client(
        user_id=ctx.user_id,
        token=ctx.auth_token,
        tenant_id=ctx.project_id,
        auth_url=auth_url
    )

    cl.management_url = auth_url

    return cl


def client_for_admin():
    return _admin_client()


def client_for_trusts(trust_id):
    return _admin_client(trust_id=trust_id)


def _admin_client(trust_id=None):
    if CONF.keystone_authtoken.auth_type is None:
        auth_url = CONF.keystone_authtoken.www_authenticate_uri
        project_name = CONF.keystone_authtoken.admin_tenant_name

        # You can't use trust and project together

        if trust_id:
            project_name = None

        cl = ks_client.Client(
            username=CONF.keystone_authtoken.admin_user,
            password=CONF.keystone_authtoken.admin_password,
            project_name=project_name,
            auth_url=auth_url,
            trusts=trust_id
        )

        cl.management_url = auth_url

        return cl
    else:
        kwargs = {}

        if trust_id:
            # Remove domain_id, domain_name, project_name and project_id,
            # since we need a trust scoped auth object
            kwargs['domain_id'] = None
            kwargs['domain_name'] = None
            kwargs['project_name'] = None
            kwargs['project_domain_name'] = None
            kwargs['project_id'] = None
            kwargs['trust_id'] = trust_id

        auth = loading.load_auth_from_conf_options(
            CONF,
            'keystone_authtoken',
            **kwargs
        )
        sess = loading.load_session_from_conf_options(
            CONF,
            'keystone',
            auth=auth
        )

        return ks_client.Client(session=sess)



# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.engine import dispatcher
from mistral.tests.unit import base
from mistral.workflow import commands


def _print_commands(cmds):
    print("commands:")

    for cmd in cmds:
        if isinstance(cmd, commands.RunTask):
            print("%s, %s, %s" % (type(cmd), cmd.is_waiting(), cmd.unique_key))
        else:
            print("%s" % type(cmd))


class CommandDispatcherTest(base.BaseTest):
    def test_rearrange_commands(self):
        no_wait = commands.RunTask(None, None, None, None)
        fail = commands.FailWorkflow(None, None, None, None)
        succeed = commands.SucceedWorkflow(None, None, None, None)

        wait1 = commands.RunTask(None, None, None, None)
        wait1.wait = True
        wait1.unique_key = 'wait1'

        wait2 = commands.RunTask(None, None, None, None)
        wait2.wait = True
        wait2.unique_key = 'wait2'

        wait3 = commands.RunTask(None, None, None, None)
        wait3.wait = True
        wait3.unique_key = 'wait3'

        # 'set state' command is the first, others must be ignored.
        initial = [fail, no_wait, wait1, wait3, wait2]
        expected = [fail]

        cmds = dispatcher._rearrange_commands(initial)

        self.assertEqual(expected, cmds)

        # 'set state' command is the last, tasks before it must be sorted.
        initial = [no_wait, wait2, wait1, wait3, succeed]
        expected = [no_wait, wait1, wait2, wait3, succeed]

        cmds = dispatcher._rearrange_commands(initial)

        self.assertEqual(expected, cmds)

        # 'set state' command is in the middle, tasks before it must be sorted
        # and the task after it must be ignored.
        initial = [wait3, wait2, no_wait, succeed, wait1]
        expected = [no_wait, wait2, wait3, succeed]

        cmds = dispatcher._rearrange_commands(initial)

        self.assertEqual(expected, cmds)



# Copyright 2015 - StackStorm, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg

from mistral.db.v2 import api as db_api
from mistral.services import workflows as wf_service
from mistral.tests.unit.engine import base
from mistral.workflow import states


# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')

MAIN_WF = """
---
version: '2.0'
main_wf:
  tasks:
    t1:
      description: Task 1
      action: std.noop
      publish:
        task1: some_value
      on-success:
        - t2_l
        - t2
        - t2_r
    t2_l:
      description: Task 2 - Left Branch
      action: std.sleep seconds=2
      publish:
        task2_l: some_value
      on-success:
        - t3_l
    t2:
      description: Task 2 - Fails here
      input:
        wrong_input: true
      workflow: sub_workflow_with_input
      publish:
        task2: some_value
      on-success:
        - t3
    t2_r:
      description: Task 2 - Right Branch
      action: std.sleep seconds=2
      publish:
        task2_r: some_value
      on-success:
        - t3_r
    t3_l:
      description: Task 3 - Left Branch
      action: std.noop
      publish:
        task3_l: some_value
    t3:
      description: Task 3 - Middle Branch
      action: std.noop
      publish:
        task3: some_value
    t3_r:
      description: Task 3 - Right Branch
      action: std.noop
      publish:
        task3_r: some_value
"""
MAIN_WITH_ITEMS_WF = """
---
version: '2.0'
main_wf:
  tasks:
    t1:
      description: Task 1
      action: std.noop
      publish:
        task1: some_value
      on-success:
        - t2_l
        - t2
        - t2_r
    t2_l:
      description: Task 2 - Left Branch
      action: std.sleep seconds=2
      publish:
        task2_l: some_value
      on-success:
        - t3_l
    t2:
      description: Task 2 - Fails here
      with-items: v in <% [1,2,3] %>
      input:
        wrong_input: true
      workflow: sub_workflow_with_input
      publish:
        task2: some_value
      on-success:
        - t3
    t2_r:
      description: Task 2 - Right Branch
      action: std.sleep seconds=2
      publish:
        task2_r: some_value
      on-success:
        - t3_r
    t3_l:
      description: Task 3 - Left Branch
      action: std.noop
      publish:
        task3_l: some_value
    t3:
      description: Task 3 - Middle Branch
      action: std.noop
      publish:
        task3: some_value
    t3_r:
      description: Task 3 - Right Branch
      action: std.noop
      publish:
        task3_r: some_value
"""
SUB_WF = """
---
version: 2.0
sub_workflow_with_input:
  input:
    - my_input
  tasks:
    sub_wf_task:
      action: std.noop
      publish:
        sub_wf_task_result: some_value
"""


class TaskGracefulFailTest(base.EngineTestCase):
    def test_graceful_fail(self):
        wf_service.create_workflows(SUB_WF)
        wf_service.create_workflows(MAIN_WF)

        wf_ex = self.engine.start_workflow('main_wf')

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

        task_1_ex = self._assert_single_item(task_execs, name='t1')
        task_2_ex = self._assert_single_item(task_execs, name='t2')
        task_2_r_ex = self._assert_single_item(task_execs, name='t2_r')
        task_2_l_ex = self._assert_single_item(task_execs, name='t2_l')
        task_3_r_ex = self._assert_single_item(task_execs, name='t3_r')
        task_3_l_ex = self._assert_single_item(task_execs, name='t3_l')

        self.assertEqual(states.SUCCESS, task_1_ex.state)
        self.assertEqual(states.ERROR, task_2_ex.state)
        self.assertIn('Invalid input', task_2_ex.state_info)
        self.assertEqual(states.SUCCESS, task_2_r_ex.state)
        self.assertEqual(states.SUCCESS, task_2_l_ex.state)
        self.assertEqual(states.SUCCESS, task_3_r_ex.state)
        self.assertEqual(states.SUCCESS, task_3_l_ex.state)

    def test_with_items_graceful_fail(self):
        wf_service.create_workflows(SUB_WF)
        wf_service.create_workflows(MAIN_WITH_ITEMS_WF)

        wf_ex = self.engine.start_workflow('main_wf')

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_execs = wf_ex.task_executions

        task_1_ex = self._assert_single_item(task_execs, name='t1')
        task_2_ex = self._assert_single_item(task_execs, name='t2')
        task_2_r_ex = self._assert_single_item(task_execs, name='t2_r')
        task_2_l_ex = self._assert_single_item(task_execs, name='t2_l')
        task_3_r_ex = self._assert_single_item(task_execs, name='t3_r')
        task_3_l_ex = self._assert_single_item(task_execs, name='t3_l')

        self.assertEqual(states.SUCCESS, task_1_ex.state)
        self.assertEqual(states.ERROR, task_2_ex.state)
        self.assertIn('Invalid input', task_2_ex.state_info)
        self.assertEqual(states.SUCCESS, task_2_r_ex.state)
        self.assertEqual(states.SUCCESS, task_2_l_ex.state)
        self.assertEqual(states.SUCCESS, task_3_r_ex.state)
        self.assertEqual(states.SUCCESS, task_3_l_ex.state)



# Copyright 2018 - Extreme Networks, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging

from mistral.tests.unit.engine import base as engine_test_base


LOG = logging.getLogger(__name__)


class NotifierTestCase(engine_test_base.EngineTestCase):
    # TODO(rakhmerov): All these method have a different signature comparing to
    # their base versions. Must be fixed.
    def await_workflow_success(self, wf_ex_id, post_delay=1):
        # Override the original wait method to add a delay to allow enough
        # time for the notification events to get processed.
        super(NotifierTestCase, self).await_workflow_success(wf_ex_id)

        self._sleep(post_delay)

    def await_workflow_error(self, wf_ex_id, post_delay=1):
        # Override the original wait method to add a delay to allow enough
        # time for the notification events to get processed.
        super(NotifierTestCase, self).await_workflow_error(wf_ex_id)

        self._sleep(post_delay)

    def await_workflow_paused(self, wf_ex_id, post_delay=1):
        # Override the original wait method to add a delay to allow enough
        # time for the notification events to get processed.
        super(NotifierTestCase, self).await_workflow_paused(wf_ex_id)

        self._sleep(post_delay)

    def await_workflow_cancelled(self, wf_ex_id, post_delay=1):
        # Override the original wait method to add a delay to allow enough
        # time for the notification events to get processed.
        super(NotifierTestCase, self).await_workflow_cancelled(wf_ex_id)

        self._sleep(post_delay)



# Copyright 2017 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.api.controllers.v2 import types
from mistral import exceptions as exc
from mistral.tests.unit import base
from mistral.utils import filter_utils


class TestTypesController(base.BaseTest):

    base_id = '88888888-4444-4444-4444-777777755555'
    uuid_type = types.uuid

    def test_uuid_type(self):
        self.uuid_type.validate(self.base_id)

    def test_uuid_type_wit_invalid_format(self):
        self.assertRaises(exc.InputException,
                          self.uuid_type.validate, 'invalid_format')
        self.assertRaises(exc.InputException,
                          self.uuid_type.validate, '44-231-454-542123')

    def test_uuid_with_filters(self):
        for filter_type in filter_utils.ALL:
            value = '{}{}'.format(filter_type + ':', self.base_id)
            if filter_type.startswith((filter_utils.IN, filter_utils.NOT_IN)):
                self.assertRaises(exc.InputException,
                                  self.uuid_type.validate, value)
            else:
                self.uuid_type.validate(value)

    def test_uuid_type_with_invalid_prefix(self):
        value = 'invalid:{}'.format(self.base_id)
        self.assertRaises(exc.InputException, self.uuid_type.validate, value)



# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc

from stevedore import extension


class Evaluator(object):
    """Expression evaluator interface.

    Having this interface gives the flexibility to change the actual expression
    language used in Mistral DSL for conditions, output calculation etc.
    """

    @classmethod
    @abc.abstractmethod
    def validate(cls, expression):
        """Parse and validates the expression.

        :param expression: Expression string
        :return: True if expression is valid
        """
        pass

    @classmethod
    @abc.abstractmethod
    def evaluate(cls, expression, context):
        """Evaluates the expression against the given data context.

        :param expression: Expression string
        :param context: Data context
        :return: Expression result
        """
        pass

    @classmethod
    @abc.abstractmethod
    def is_expression(cls, expression):
        """Check expression string and decide whether it is expression or not.

        :param expression: Expression string
        :return: True if string is expression
        """
        pass


def get_custom_functions():
    """Get custom functions.

    Retrieves the list of custom functions used in YAQL/Jinja expressions.
    """
    # {name => function object).
    result = dict()

    mgr = extension.ExtensionManager(
        namespace='mistral.expression.functions',
        invoke_on_load=False
    )

    for name in mgr.names():
        result[name] = mgr[name].plugin

    return result



# Copyright 2020 Nokia Software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Delete delayed calls with key=NULL.

Revision ID: 038
Revises: 037
Create Date: 2020-7-13 13:20:00

"""

# revision identifiers, used by Alembic.

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import table, column

revision = '038'
down_revision = '037'


def upgrade():
    # See https://bugs.launchpad.net/mistral/+bug/1861988.
    # Due to this bug there may be redundant delayed calls in DB.
    # We need to delete all rows where the "key" column is None.
    session = sa.orm.Session(bind=op.get_bind())

    delayed_calls = table('delayed_calls_v2', column('key'))

    with session.begin(subtransactions=True):
        session.execute(
            delayed_calls.delete().where(delayed_calls.c.key==None)  # noqa
        )

    session.commit()



#    Copyright 2016 NEC Corporation.  All rights reserved.
#    Copyright 2019 Nokia Software.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from dateutil import parser as dateparser


EQUALS = 'eq'
NOT_EQUAL = 'neq'
LESS_THAN = 'lt'
LESS_THAN_EQUALS = 'lte'
GREATER_THAN = 'gt'
GREATER_THAN_EQUALS = 'gte'
IN = 'in'
NOT_IN = 'nin'
HAS = 'has'

ALL = (
    EQUALS,
    NOT_EQUAL,
    GREATER_THAN,
    GREATER_THAN_EQUALS,
    LESS_THAN,
    LESS_THAN_EQUALS,
    HAS,
    IN,
    NOT_IN
)


def create_filters_from_request_params(none_values=None, **params):
    """Create filters from REST request parameters.

    :param none_values: field names, where the value is required to be None.
    :param req_params: REST request parameters.
    :return: filters dictionary.
    """
    none_values = none_values or []
    filters = {}

    for column, data in params.items():
        if (data is None and column in none_values) or data is not None:
            if isinstance(data, str):
                f_type, value = extract_filter_type_and_value(data)

                create_or_update_filter(column, value, f_type, filters)
            else:
                create_or_update_filter(column, data, _filter=filters)

    return filters


def create_or_update_filter(column, value, filter_type='eq', _filter=None):
    """Create or Update filter.

    :param column: Column name by which user want to filter.
    :param value: Column value.
    :param filter_type: filter type. Filter type can be
                        'eq', 'neq', 'gt', 'gte', 'lte', 'in',
                        'lt', 'nin'. Default is 'eq'.
    :param _filter: Optional. If provided same filter dictionary will
                    be updated.
    :return: filter dictionary.

    """
    if _filter is None:
        _filter = {}

    _filter[column] = {filter_type: value}

    return _filter


def extract_filter_type_and_value(data):
    """Extract filter type and its value from the data.

    :param data: REST parameter value from which filter type and
                 value can be get. It should be in format of
                 'filter_type:value'.
    :return: filter type and value.
    """
    if has_filters(data):
        filter_type, value = data.split(':', 1)
        value = str(value)
        if data.startswith((IN, NOT_IN)):
            value = list(value.split(","))
    else:
        value = data
        filter_type = EQUALS

    return filter_type, value


def has_filters(value):
    for filter_type in ALL:
        if value.startswith(filter_type + ':'):
            return True

    return False


def match_filter(obj, attr_name, attr_filter):
    # If the attribute doesn't exist we assume that any filter is
    # not applicable and we ignore it.
    if not hasattr(obj, attr_name):
        return True

    attr_val = getattr(obj, attr_name)

    for op, val in attr_filter.items():
        # If the attribute is a date and the given filter value is a string
        # we try to convert the filter value into a data as well.
        if isinstance(attr_val, datetime.datetime) and isinstance(val, str):
            val = dateparser.isoparse(val)

        if op not in ALL:
            raise ValueError(
                'Unknown filter operation encountered [operation=%s]' % op
            )

        if op == EQUALS and attr_val != val:
            return False

        if op == NOT_EQUAL and attr_val == val:
            return False

        if op == LESS_THAN and attr_val >= val:
            return False

        if op == LESS_THAN_EQUALS and attr_val > val:
            return False

        if op == GREATER_THAN and attr_val <= val:
            return False

        if op == GREATER_THAN_EQUALS and attr_val < val:
            return False

        if op == IN and attr_val not in val:
            return False

        if op == NOT_IN and attr_val in val:
            return False

        if op == HAS and val not in attr_val:
            return False

    return True


def match_filters(obj, filters):
    for attr_name, attr_filter in filters.items():
        if not match_filter(obj, attr_name, attr_filter):
            return False

    return True



# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - Huawei Technologies Co. Ltd
# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import contextlib
import inspect
import os
import shutil
import tempfile
import threading

from oslo_concurrency import processutils
from oslo_serialization import jsonutils

from mistral_lib.utils import inspect_utils

from mistral import exceptions as exc
from mistral import expressions as expr


# Thread local storage.
_th_loc_storage = threading.local()


@contextlib.contextmanager
def tempdir(**kwargs):
    argdict = kwargs.copy()

    if 'dir' not in argdict:
        argdict['dir'] = '/tmp/'

    tmpdir = tempfile.mkdtemp(**argdict)

    try:
        yield tmpdir
    finally:
        try:
            shutil.rmtree(tmpdir)
        except OSError as e:
            raise exc.DataAccessException(
                "Failed to delete temp dir %(dir)s (reason: %(reason)s)" %
                {'dir': tmpdir, 'reason': e}
            )


def save_text_to(text, file_path, overwrite=False):
    if os.path.exists(file_path) and not overwrite:
        raise exc.DataAccessException(
            "Cannot save data to file. File %s already exists."
        )

    with open(file_path, 'w') as f:
        f.write(text)


def generate_key_pair(key_length=2048):
    """Create RSA key pair with specified number of bits in key.

    Returns tuple of private and public keys.
    """
    with tempdir() as tmpdir:
        keyfile = os.path.join(tmpdir, 'tempkey')
        args = [
            'ssh-keygen',
            '-q',  # quiet
            '-N', '',  # w/o passphrase
            '-t', 'rsa',  # create key of rsa type
            '-f', keyfile,  # filename of the key file
            '-C', 'Generated-by-Mistral'  # key comment
        ]

        if key_length is not None:
            args.extend(['-b', key_length])

        processutils.execute(*args)

        if not os.path.exists(keyfile):
            raise exc.DataAccessException(
                "Private key file hasn't been created"
            )

        private_key = open(keyfile).read()
        public_key_path = keyfile + '.pub'

        if not os.path.exists(public_key_path):
            raise exc.DataAccessException(
                "Public key file hasn't been created"
            )
        public_key = open(public_key_path).read()

        return private_key, public_key


def to_json_str(obj):
    """Serializes an object into a JSON string.

    :param obj: Object to serialize.
    :return: JSON string.
    """

    if obj is None:
        return None

    def _fallback(value):
        if inspect.isgenerator(value):
            result = list(value)

            # The result of the generator call may be again not primitive
            # so we need to call "to_primitive" again with the same fallback
            # function. Note that the endless recursion here is not a problem
            # because "to_primitive" limits the depth for custom classes,
            # if they are present in the object graph being traversed.
            return jsonutils.to_primitive(
                result,
                convert_instances=True,
                fallback=_fallback
            )

        return value

    # We need to convert the root of the given object graph into
    # a primitive by hand so that we also enable conversion of
    # object of custom classes into primitives. Otherwise, they are
    # ignored by the "json" lib.
    return jsonutils.dumps(
        jsonutils.to_primitive(obj, convert_instances=True, fallback=_fallback)
    )


def from_json_str(json_str):
    """Reconstructs an object from a JSON string.

    :param json_str: A JSON string.
    :return: Deserialized object.
    """

    if json_str is None:
        return None

    return jsonutils.loads(json_str)


def evaluate_object_fields(obj, ctx):
    """Evaluates all expressions recursively contained in the object fields.

    Some of the given object fields may be strings or data structures that
    contain YAQL/Jinja expressions. The method evaluates them and updates
    the corresponding object fields with the evaluated values.

    :param obj: The object to inspect.
    :param ctx: Expression context.
    """
    fields = inspect_utils.get_public_fields(obj)

    evaluated_fields = expr.evaluate_recursively(fields, ctx)

    for k, v in evaluated_fields.items():
        setattr(obj, k, v)



# Copyright 2018 - Extreme Networks, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging

from mistral.notifiers import base


LOG = logging.getLogger(__name__)


class NoopPublisher(base.NotificationPublisher):

    def publish(self, ctx, ex_id, data, event, timestamp, **kwargs):
        LOG.info(
            'The event %s for [name=%s, id=%s] is published by the '
            'noop notification publisher.',
            event,
            data.get('name'),
            ex_id
        )



# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
from unittest import mock

import requests

from mistral.actions import std_actions as std
from mistral.tests.unit import base
from mistral_lib import actions as mistral_lib_actions


URL = 'http://some_url'

DATA = {
    'server': {
        'id': '12345',
        'metadata': {
            'name': 'super_server'
        }
    }
}

ACTION_LOGGER = 'mistral.actions.std_actions'


def get_fake_response(content, code, **kwargs):
    return base.FakeHTTPResponse(
        content,
        code,
        **kwargs
    )


def get_success_fake_response():
    return get_fake_response(
        json.dumps(DATA),
        200,
        headers={'Content-Type': 'application/json'}
    )


def get_error_fake_response():
    return get_fake_response(
        json.dumps(DATA),
        401
    )


class HTTPActionTest(base.BaseTest):
    @mock.patch.object(requests, 'request')
    def test_http_action(self, mocked_method):
        mocked_method.return_value = get_success_fake_response()
        mock_ctx = mock.Mock()

        action = std.HTTPAction(
            url=URL,
            method='POST',
            body=DATA,
            timeout=20,
            allow_redirects=True
        )

        DATA_STR = json.dumps(DATA)

        self.assertEqual(DATA_STR, action.body)
        self.assertEqual(URL, action.url)

        result = action.run(mock_ctx)

        self.assertIsInstance(result, dict)
        self.assertEqual(DATA, result['content'])
        self.assertIn('headers', result)
        self.assertEqual(200, result['status'])

        mocked_method.assert_called_with(
            'POST',
            URL,
            data=DATA_STR,
            json=None,
            headers=None,
            cookies=None,
            params=None,
            timeout=20,
            auth=None,
            allow_redirects=True,
            proxies=None,
            verify=None
        )

    @mock.patch.object(requests, 'request')
    def test_http_action_error_result(self, mocked_method):
        mocked_method.return_value = get_error_fake_response()
        mock_ctx = mock.Mock()

        action = std.HTTPAction(
            url=URL,
            method='POST',
            body=DATA,
            timeout=20,
            allow_redirects=True
        )

        result = action.run(mock_ctx)

        self.assertIsInstance(result, mistral_lib_actions.Result)
        self.assertEqual(401, result.error['status'])

    @mock.patch.object(requests, 'request')
    def test_http_action_with_auth(self, mocked_method):
        mocked_method.return_value = get_success_fake_response()
        mock_ctx = mock.Mock()

        action = std.HTTPAction(
            url=URL,
            method='POST',
            auth='user:password'
        )

        action.run(mock_ctx)

        args, kwargs = mocked_method.call_args
        self.assertEqual(('user', 'password'), kwargs['auth'])

    @mock.patch.object(requests, 'request')
    def test_http_action_with_headers(self, mocked_method):
        mocked_method.return_value = get_success_fake_response()
        mock_ctx = mock.Mock()

        headers = {'int_header': 33, 'bool_header': True,
                   'float_header': 3.0, 'regular_header': 'teststring'}

        safe_headers = {'int_header': '33', 'bool_header': 'True',
                        'float_header': '3.0', 'regular_header': 'teststring'}

        action = std.HTTPAction(
            url=URL,
            method='POST',
            headers=headers.copy(),
        )

        result = action.run(mock_ctx)

        self.assertIn('headers', result)

        args, kwargs = mocked_method.call_args
        self.assertEqual(safe_headers, kwargs['headers'])

    @mock.patch.object(requests, 'request')
    def test_http_action_empty_resp(self, mocked_method):

        def invoke(content):
            action = std.HTTPAction(
                url=URL,
                method='GET',
            )
            mocked_method.return_value = get_fake_response(
                content=content, code=200
            )
            result = action.run(mock.Mock())
            self.assertEqual(content, result['content'])

        invoke(None)
        invoke('')

    @mock.patch.object(requests, 'request')
    def test_http_action_none_encoding_not_empty_resp(self, mocked_method):
        action = std.HTTPAction(
            url=URL,
            method='GET',
        )

        mocked_method.return_value = get_fake_response(
            content='', code=200, encoding=None
        )
        mock_ctx = mock.Mock()
        result = action.run(mock_ctx)

        self.assertIsNone(result['encoding'])

    @mock.patch.object(requests, 'request')
    def test_http_action_hides_request_body_if_needed(self, mocked_method):
        self.override_config(
            'hide_request_body',
            True,
            group='action_logging'
        )

        sensitive_data = 'I actually love anime.'

        action = std.HTTPAction(url=URL, method='POST', body=sensitive_data)

        mocked_method.return_value = get_fake_response(
            content='', code=201
        )
        mock_ctx = mock.Mock()

        with self.assertLogs(logger=ACTION_LOGGER, level='INFO') as logs:
            action.run(mock_ctx)

        self.assertEqual(2, len(logs.output))  # Request and response loglines

        log = logs.output[0]  # Request log
        msg = "Request body hidden due to action_logging configuration."
        self.assertNotIn(sensitive_data, log)
        self.assertIn(msg, log)

    @mock.patch.object(requests, 'request')
    def test_http_action_hides_request_headers_if_needed(self, mocked_method):
        sensitive_header = 'Authorization'
        self.override_config(
            'sensitive_headers',
            [sensitive_header],
            group='action_logging'
        )

        headers = {
            sensitive_header: 'Bearer 13e7aa3fc23e50bc1529dc136791d34d'
        }

        action = std.HTTPAction(url=URL, method='GET', headers=headers)

        mocked_method.return_value = get_fake_response(
            content='', code=200
        )
        mock_ctx = mock.Mock()

        with self.assertLogs(logger=ACTION_LOGGER, level='INFO') as logs:
            action.run(mock_ctx)

        self.assertEqual(2, len(logs.output))  # Request and response loglines

        log = logs.output[0]  # Request log
        self.assertNotIn(headers[sensitive_header], log)
        self.assertIn('headers={}', log)

    @mock.patch.object(requests, 'request')
    def test_http_action_hides_response_body_if_needed(self, mocked_method):
        self.override_config(
            'hide_response_body',
            True,
            group='action_logging'
        )

        action = std.HTTPAction(url=URL, method='GET')

        sensitive_data = 'I actually love anime.'

        mocked_method.return_value = get_fake_response(
            content=sensitive_data, code=200
        )
        mock_ctx = mock.Mock()

        with self.assertLogs(logger=ACTION_LOGGER, level='INFO') as logs:
            action.run(mock_ctx)

        self.assertEqual(2, len(logs.output))  # Request and response loglines

        log = logs.output[1]  # Response log
        msg = "Response body hidden due to action_logging configuration."
        self.assertNotIn(sensitive_data, log)
        self.assertIn(msg, log)



# Copyright 2015 - Mirantis, Inc.
# Copyright 2020 - Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc

from mistral import config as cfg
from mistral import exceptions as exc
from mistral import utils

from oslo_utils import importutils
from stevedore import driver
from stevedore import extension

_PYV8 = importutils.try_import('PyV8')
_V8EVAL = importutils.try_import('v8eval')
_PY_MINI_RACER = importutils.try_import('py_mini_racer.py_mini_racer')
_EVALUATOR = None


class JSEvaluator(object):
    @classmethod
    @abc.abstractmethod
    def evaluate(cls, script, context):
        """Executes given JavaScript.

        :param script: The text of JavaScript snippet that needs to be
        executed.
        :param context: This object will be assigned to the $ javascript
        variable.
        :return result of evaluated javascript code.
        :raise MistralException: if corresponding js library is not installed.
        """
        pass


class PyV8Evaluator(JSEvaluator):
    @classmethod
    def evaluate(cls, script, ctx):
        if not _PYV8:
            raise exc.MistralException(
                "PyV8 module is not available. Please install PyV8."
            )

        with _PYV8.JSContext() as js_ctx:
            # Prepare data context and way for interaction with it.
            js_ctx.eval('$ = %s' % utils.to_json_str(ctx))

            result = js_ctx.eval(script)

            return _PYV8.convert(result)


class V8EvalEvaluator(JSEvaluator):
    @classmethod
    def evaluate(cls, script, ctx):
        if not _V8EVAL:
            raise exc.MistralException(
                "v8eval module is not available. Please install v8eval."
            )

        v8 = _V8EVAL.V8()

        ctx_str = utils.to_json_str(ctx)

        return v8.eval(
            '$ = %s; %s' % (ctx_str, script)
        )


class PyMiniRacerEvaluator(JSEvaluator):
    @classmethod
    def evaluate(cls, script, ctx):
        if not _PY_MINI_RACER:
            raise exc.MistralException(
                "PyMiniRacer module is not available. Please install "
                "PyMiniRacer."
            )

        js_ctx = _PY_MINI_RACER.MiniRacer()

        return js_ctx.eval(
            '$ = {}; {}'.format(utils.to_json_str(ctx), script)
        )


_mgr = extension.ExtensionManager(
    namespace='mistral.expression.evaluators',
    invoke_on_load=False
)


def get_js_evaluator():
    global _EVALUATOR

    if not _EVALUATOR:
        mgr = driver.DriverManager(
            'mistral.js.implementation',
            cfg.CONF.js_implementation,
            invoke_on_load=True
        )

        _EVALUATOR = mgr.driver

    return _EVALUATOR


def evaluate(script, ctx):
    return get_js_evaluator().evaluate(script, ctx)



# Copyright 2018 OVH SAS. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models
from mistral.services import adhoc_actions
from mistral.tests.unit.api import base
from mistral.tests.unit.mstrlfixtures import policy_fixtures


MOCK_DELETE = mock.MagicMock(return_value=None)

ADHOC_ACTION_YAML = """
---
version: '2.0'

my_action:
  description: My super cool action.
  tags: ['test', 'v2']
  base: std.echo
  base-input:
    output: "{$.str1}{$.str2}"
"""

ADHOC_ACTION_DEF = models.ActionDefinition(
    id='123e4567-e89b-12d3-a456-426655440000',
    name='my_action',
    is_system=False,
    description='My super cool action.',
    tags=['test', 'v2'],
    definition=ADHOC_ACTION_YAML
)

MOCK_ACTION = mock.MagicMock(return_value=ADHOC_ACTION_DEF)


class TestActionPolicy(base.APITest):
    """Test action related policies

    Policies to test:
    - actions:create
    - actions:delete
    - actions:get
    - actions:list
    - actions:publicize (on POST & PUT)
    - actions:update
    """

    def setUp(self):
        super(TestActionPolicy, self).setUp()

        self.policy = self.useFixture(policy_fixtures.PolicyFixture())

    @mock.patch.object(db_api, "create_action_definition")
    def test_action_create_not_allowed(self, mock_obj):
        self.policy.change_policy_definition(
            {"actions:create": "role:FAKE"}
        )

        resp = self.app.post(
            '/v2/actions',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    @mock.patch.object(db_api, "create_action_definition")
    def test_action_create_allowed(self, mock_obj):
        self.policy.change_policy_definition(
            {"actions:create": "role:FAKE or rule:admin_or_owner"}
        )

        resp = self.app.post(
            '/v2/actions',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(201, resp.status_int)

    @mock.patch.object(db_api, "create_action_definition")
    def test_action_create_public_not_allowed(self, mock_obj):
        self.policy.change_policy_definition({
            "actions:create": "role:FAKE or rule:admin_or_owner",
            "actions:publicize": "role:FAKE"
        })

        resp = self.app.post(
            '/v2/actions?scope=public',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    @mock.patch.object(db_api, "create_action_definition")
    def test_action_create_public_allowed(self, mock_obj):
        self.policy.change_policy_definition({
            "actions:create": "role:FAKE or rule:admin_or_owner",
            "actions:publicize": "role:FAKE or rule:admin_or_owner"
        })

        resp = self.app.post(
            '/v2/actions?scope=public',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(201, resp.status_int)

    @mock.patch.object(db_api, "delete_action_definition", MOCK_DELETE)
    @mock.patch.object(db_api, "get_action_definition", MOCK_ACTION)
    def test_action_delete_not_allowed(self):
        self.policy.change_policy_definition(
            {"actions:delete": "role:FAKE"}
        )

        resp = self.app.delete('/v2/actions/123', expect_errors=True)

        self.assertEqual(403, resp.status_int)

    @mock.patch.object(db_api, "delete_action_definition", MOCK_DELETE)
    @mock.patch.object(db_api, "get_action_definition", MOCK_ACTION)
    def test_action_delete_allowed(self):
        self.policy.change_policy_definition(
            {"actions:delete": "role:FAKE or rule:admin_or_owner"}
        )

        resp = self.app.delete('/v2/actions/123', expect_errors=True)

        self.assertEqual(204, resp.status_int)

    def test_action_get_not_allowed(self):
        self.policy.change_policy_definition(
            {"actions:get": "role:FAKE"}
        )

        resp = self.app.get('/v2/actions/123', expect_errors=True)

        self.assertEqual(403, resp.status_int)

    def test_action_get_allowed(self):
        # Create an adhoc action for the purpose of the test.
        adhoc_actions.create_actions(ADHOC_ACTION_YAML)

        self.policy.change_policy_definition(
            {"actions:get": "role:FAKE or rule:admin_or_owner"}
        )

        resp = self.app.get('/v2/actions/my_action')

        self.assertEqual(200, resp.status_int)

    def test_action_list_not_allowed(self):
        self.policy.change_policy_definition(
            {"actions:list": "role:FAKE"}
        )

        resp = self.app.get('/v2/actions', expect_errors=True)

        self.assertEqual(403, resp.status_int)

    def test_action_list_allowed(self):
        self.policy.change_policy_definition(
            {"actions:list": "role:FAKE or rule:admin_or_owner"}
        )

        resp = self.app.get('/v2/actions', expect_errors=True)

        self.assertEqual(200, resp.status_int)

    @mock.patch.object(db_api, "update_action_definition")
    def test_action_update_not_allowed(self, mock_obj):
        self.policy.change_policy_definition(
            {"actions:update": "role:FAKE"}
        )

        resp = self.app.put(
            '/v2/actions',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    @mock.patch.object(db_api, "update_action_definition")
    def test_action_update_allowed(self, mock_obj):
        self.policy.change_policy_definition(
            {"actions:update": "role:FAKE or rule:admin_or_owner"}
        )

        resp = self.app.put(
            '/v2/actions',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)

    @mock.patch.object(db_api, "update_action_definition")
    def test_action_update_public_not_allowed(self, mock_obj):
        self.policy.change_policy_definition({
            "actions:update": "role:FAKE or rule:admin_or_owner",
            "actions:publicize": "role:FAKE"
        })

        resp = self.app.put(
            '/v2/actions?scope=public',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    @mock.patch.object(db_api, "update_action_definition")
    def test_action_update_public_allowed(self, mock_obj):
        self.policy.change_policy_definition({
            "actions:update": "role:FAKE or rule:admin_or_owner",
            "actions:publicize": "role:FAKE or rule:admin_or_owner"
        })

        resp = self.app.put(
            '/v2/actions?scope=public',
            ADHOC_ACTION_YAML,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)



#  Copyright 2019 - Nokia Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import eventlet
from eventlet import semaphore
from unittest import mock

from mistral.api.controllers.v2 import execution
from mistral import context
from mistral.db.v2 import api as db_api
from mistral.services import workflows as wf_service
from mistral.tests.unit.api import base
from mistral.tests.unit import base as unit_base
from mistral.tests.unit.engine import base as engine_base


WF_TEXT = """---
version: '2.0'

wf:
  tasks:
       task1:
         action: std.noop
"""


class TestParallelOperations(base.APITest, engine_base.EngineTestCase):
    def setUp(self):
        super(TestParallelOperations, self).setUp()

        wf_service.create_workflows(WF_TEXT)

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        self.wf_ex_id = wf_ex.id

        self.decorator_call_cnt = 0
        self.threads = []
        self.addCleanup(self.kill_threads)

    def kill_threads(self):
        for thread in self.threads:
            thread.kill()

    def test_parallel_api_list_and_delete_operations(self):
        # One execution already exists. Let's create another one.
        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        self.assertEqual(2, len(db_api.get_workflow_executions()))

        delete_lock = semaphore.Semaphore(0)
        list_lock = semaphore.Semaphore(0)

        orig_func = execution._get_workflow_execution_resource

        def delete_():
            context.set_ctx(unit_base.get_context())

            db_api.delete_workflow_execution(self.wf_ex_id)

            # Unlocking the "list" operation.
            list_lock.release()

        def list_():
            resp = self.app.get('/v2/executions/')

            self.assertEqual(1, len(resp.json['executions']))

        # This decorator is needed to halt the thread of the "list"
        # operation and wait till the "delete" operation is over.
        # That way we'll reproduce the situation when the "list"
        # operation has already fetched the execution but it then
        # gets deleted before further lazy-loading of the execution
        # fields.
        def decorate_resource_function_(arg):
            self.decorator_call_cnt += 1

            # It makes sense to use this trick only once since only
            # one object gets deleted.
            if self.decorator_call_cnt == 1:
                # It's OK now to delete the execution so we release
                # the corresponding lock.
                delete_lock.release()

                # Wait till the "delete" operation has finished.
                list_lock.acquire()

            return orig_func(arg)

        with mock.patch.object(execution, '_get_workflow_execution_resource',
                               wraps=decorate_resource_function_):
            self.threads.append(eventlet.spawn(list_))

            # Make sure that the "list" operation came to the right point
            # which is just about the call to the resource function.
            delete_lock.acquire()

            self.threads.append(eventlet.spawn(delete_))

        for t in self.threads:
            t.wait()



# Copyright 2019 - Nokia Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg


def is_validation_enabled(validate):
    validation_mode = cfg.CONF.api.validation_mode

    if validation_mode == 'mandatory':
        result = True
    elif validation_mode == 'disabled':
        result = False
    else:
        # validation_mode = 'enabled'
        result = validate

    return result
